import collections
import enum
import os

from six import iteritems, iterkeys, itervalues
from six.moves import cStringIO as StringIO

import ducky.cpu
import ducky.util

def dump_node(node):
  line = None

  if node.coord is not None and node.coord.line != 0:
    with open(node.coord.file, 'r') as f:
      line = f.readlines()[node.coord.line - 1].strip()

  ignore = ['attr_names', 'children', 'show']
  return '%s(%s)%s' % (node.__class__.__name__, ', '.join(['%s=%s' % (name, getattr(node, name)) for name in dir(node) if not name.startswith('__') and name not in ignore]), ' [' + line + ']' if line is not None else '')

def show_node(node):
  buff = StringIO()
  node.show(buf = buff, attrnames = True)
  return buff.getvalue().strip()

class CompilerError(Exception):
  def __init__(self, location, msg):
    super(CompilerError, self).__init__('%s:%s: %s' % (os.path.abspath(location.file), location.line, msg))

class SymbolAlreadyDefinedError(CompilerError):
  def __init__(self, loc, symbol):
    super(SymbolAlreadyDefinedError, self).__init__(loc, 'Symbol already defined: name=%s' % symbol)

class SymbolConflictError(CompilerError):
  pass

class SymbolUndefined(CompilerError):
  def __init__(self, loc, symbol):
    super(SymbolUndefined, self).__init__(loc, 'Undefined symbol: name=%s' % symbol)

class IncompatibleTypesError(CompilerError):
  def __init__(self, loc, t1, t2):
    super(IncompatibleTypesError, self).__init__(loc, 'Incompatible types: "%s" cannot be matched with "%s"' % (t1, t2))

class UnableToImplicitCastError(CompilerError):
  def __init__(self, loc, t1, t2):
    super(UnableToImplicitCastError, self).__init__(loc, 'Unable to perform implicit cast: "%s" => "%s"' % (t1, t2))

class NotAPointerError(CompilerError):
  def __init__(self, loc, t):
    super(NotAPointerError, self).__init__(loc, 'Type is not a pointer: "%s"' % t)

class IsAPointerError(CompilerError):
  def __init__(self, loc, t):
    super(IsAPointerError, self).__init__(loc, 'Type is a pointer: "%s"' % t)

class UndefinedStructMemberError(CompilerError):
  def __init__(self, loc, s, m):
    super(UndefinedStructMemberError, self).__init__(loc, 'Unknown structure member: "%s.%s"' % (s, m))

class Symbol(object):
  def __init__(self, visitor, name, decl_type, extern = False, defined = False, const = False):
    self.visitor = visitor
    self.name = name
    self.type = decl_type

    self.extern = extern
    self.defined = defined
    self.const = const

    self.storage = None

  def __repr__(self):
    return '<Symbol: name=%s, type=%s, storage=%s, extern=%s, defined=%s, const=%s>' % (self.name, self.type, repr(self.storage) if self.storage is not None else '<none>', self.extern, self.defined, self.const)


class SymbolStorage(object):
  def __init__(self, symbol, register = None):
    self.symbol = symbol

    self.register = register

  def __repr__(self):
    return '<%s: symbol=%s, register=%s>' % (self.__class__.__name__, self.symbol.name if self.symbol is not None else '<none>', self.register.name if self.register is not None else '<none>')

  def addrof(self, reg, emit):
    raise NotImplementedError('Storage %s does not provide addrof operator' % self.__class__.__name__)

  def name(self):
    raise NotImplementedError('Storage %s does not provide name operator' % self.__class__.__name__)

  def has_register(self):
    return self.register is not None

  def acquire_register(self, register):
    assert self.register is None
    assert register.storage is None

    self.register = register
    register.storage = self

  def release_register(self):
    assert self.register is not None

    self.register.storage = None
    self.register = None

  def spill_register(self, visitor):
    visitor.DEBUG(visitor.log_prefix + 'spill_register: reg=%s', self.register)

    if not self.has_register():
      return

    if self.register.dirty is not True:
      return

    if len(self.symbol.type) == 1:
      visitor.EMIT(STB(self.name(), self.register.name))

    elif len(self.symbol.type) == 2:
      visitor.EMIT(STS(self.name(), self.register.name))

    elif len(self.symbol.type) == 4:
      visitor.EMIT(STW(self.name(), self.register.name))

    else:
      visitor.WARN(visitor.log_prefix + 'Unhandled spill size branch')

    self.register.dirty = False

  def unspill_register(self, visitor, register):
    if self.register is not None:
      visitor.WARN(visitor.log_prefix + 'Storage already has a register!')
      return

    self.acquire_register(register)
    self.register.dirty = False

    if len(self.symbol.type) == 1:
      visitor.EMIT(LB(self.register.name, self.name()))

    elif len(self.symbol.type) == 2:
      visitor.EMIT(LS(self.register.name, self.name()))

    elif len(self.symbol.type) == 4:
      visitor.EMIT(LW(self.register.name, self.name()))

    else:
      visitor.WARN(visitor.log_prefix + 'Unhandled unspill size branch')

class StackSlotStorage(SymbolStorage):
  def __init__(self, symbol, offset):
    super(StackSlotStorage, self).__init__(symbol)

    self.offset = offset

  def addrof(self, register, emit):
    emit(MOV(register.name, 'fp'))
    emit(ADD(register.name, self.offset))

  def name(self):
    return 'fp[%i]' % self.offset

class MemorySlotStorage(SymbolStorage):
  def __init__(self, symbol, label):
    super(MemorySlotStorage, self).__init__(symbol)

    self.label = '&%s' % label

  def addrof(self, register, emit):
    emit(LA(register.name, self.label))

  def name(self):
    return self.label

class Scope(object):
  scope_id = 0

  def __init__(self, visitor, parent = None):
    self.visitor = visitor

    self.id = Scope.scope_id
    Scope.scope_id += 1

    self.parent = parent

    self.symbols = {}

  def __repr__(self):
    return '<Scope: id=%i>' % self.id

  def add(self, loc, symbol):
    self.visitor.DEBUG(self.visitor.log_prefix + 'scope: add symbol: scope=%s, symbol=%s', self, symbol)

    old_symbol = self.symbols.get(symbol.name)

    if old_symbol is not None:
      if old_symbol.defined:
        if symbol.defined:
          raise SymbolAlreadyDefinedError(loc, symbol.name)

        else:
          return old_symbol

      else:
        self.symbols[symbol.name] = symbol
        return symbol

    self.symbols[symbol.name] = symbol
    return symbol

  def get(self, name):
    self.visitor.DEBUG(self.visitor.log_prefix + 'scope: get symbol: scope=%s, symbol=%s', self, name)

    if name in self.symbols:
      return self.symbols[name]

    if self.parent is not None:
      return self.parent.get(name)

    return None


class Register(object):
  def __init__(self, rset, index):
    self.set = rset
    self.index = index

    self.name = 'r%i' % index

    self.visitor = self.set.visitor

    self.storage = None
    self.dirty = False

  def __repr__(self):
    return '<%s: storage=%s, dirty=%s>' % (self.name, self.storage.name() if self.storage is not None else '<none>', self.dirty)

  def put(self):
    self.visitor.DEBUG(self.visitor.log_prefix + 'Register.put: reg=%s', self)

    if self.storage is None:
      self.set.free_regs[self.index] = self

      self.visitor.EMIT(Comment('reg %s released' % self.name))

    else:
      self.visitor.DEBUG(self.visitor.log_prefix + 'Register.put: register acquired by storage, not free')

    self.visitor.DEBUG(self.visitor.log_prefix + 'free regs: %s', list(iterkeys(self.set.free_regs)))

  def free(self):
    self.visitor.DEBUG(self.visitor.log_prefix + 'Register.free: reg=%s', self)
    self.visitor.DEBUG(self.visitor.log_prefix + '  free regs: %s', list(iterkeys(self.set.free_regs)))

    if self.storage is not None:
      self.storage.spill_register(self.visitor)
      self.storage.release_register()

    self.set.free_regs[self.index] = self

    self.visitor.EMIT(Comment('reg %s forcefully freed' % self.name))
    return self

class RegisterSet(object):
  def __init__(self, fn):
    self.fn = fn
    self.visitor = fn.visitor

    self.all_regs = collections.OrderedDict([(r.value, Register(self, r.value)) for r in ducky.cpu.registers.GENERAL_REGISTERS])
    self.free_regs = self.all_regs.copy()
    self.used_regs = collections.OrderedDict()

    self.callee_saved_regs = self.all_regs.copy()

    self.context_stack = []

    self.DEBUG = self.visitor.DEBUG

  def save_callee_saves(self, block):
    self.DEBUG(self.visitor.log_prefix + 'RegisterSet.save_callee_saves: used_regs=%s, callee_saves=%s', sorted([r.index for r in iterkeys(self.used_regs)]), sorted(list(iterkeys(self.callee_saved_regs))))

    for reg in sorted(list(iterkeys(self.used_regs)), key = lambda x: x.index):
      if reg.index not in self.callee_saved_regs:
        continue

      self.DEBUG(self.visitor.log_prefix + '  save %s', reg.name)
      block.emit(PUSH(reg.name))

  def restore_callee_saves(self, block):
    self.DEBUG(self.visitor.log_prefix + 'RegisterSet.restore_callee_saves: used_regs=%s, callee_saves=%s', sorted([r.index for r in iterkeys(self.used_regs)]), sorted(list(iterkeys(self.callee_saved_regs))))

    for reg in reversed(sorted(list(iterkeys(self.used_regs)), key = lambda x: x.index)):
      if reg.index not in self.callee_saved_regs:
        continue

      block.emit(POP(reg.name))

  def get(self, preferred = None, keep = None):
    self.DEBUG(self.visitor.log_prefix + 'RegisterSet.get: preferred=%s, keep=%s', preferred, keep)
    self.DEBUG(self.visitor.log_prefix + '  free regs: %s', list(iterkeys(self.free_regs)))

    if self.free_regs:
      selected = None

      if preferred is not None:
        for i, reg in iteritems(self.free_regs):
          self.DEBUG(self.visitor.log_prefix + '    consider %i: %s', i, reg)
          if i != preferred and reg != preferred:
            continue

          selected = reg
          break

      else:
        selected = self.free_regs[sorted(list(iterkeys(self.free_regs)))[-1]]

      self.DEBUG(self.visitor.log_prefix + '  selected reg: %s', selected)

      if selected is not None:
        del self.free_regs[selected.index]
        self.used_regs[selected] = True
        self.visitor.EMIT(Comment('reg %s acquired' % selected.name))
        return selected

    return self.__spoil_any_reg(keep = keep)

  def __enter__(self, *args, **kwargs):
    r = self.get(*args, **kwargs)
    self.context_stack.append(r)
    return r

  def __exit__(self, *args, **kwargs):
    R = self.context_stack.pop()
    R.free()

from ducky.cpu.instructions import DuckyOpcodes, DuckyInstructionSet

class Instruction(object):
  def __init__(self, opcode, *operands):
    self.opcode = opcode
    self.operands = operands

    for inst in DuckyInstructionSet.instructions:
      if inst.opcode != opcode:
        continue

      self.mnemonic = inst.mnemonic
      break

    self.labels = []

  def __repr__(self):
    return self.materialize()

  def materialize(self):
    if self.operands:
      return '%s %s' % (self.mnemonic, ', '.join([str(o) for o in self.operands]))
    else:
      return self.mnemonic

class Directive(Instruction):
  def __init__(self, directive):
    super(Directive, self).__init__(None)

    self.directive = directive

  def materialize(self):
    return self.directive

class Comment(object):
  def __init__(self, comment):
    self.comment = comment

  def materialize(self):
    return '; %s' % self.comment

class InlineAsm(Instruction):
  def __init__(self, code):
    super(InlineAsm, self).__init__(None)

    self.code = code

  def materialize(self):
    return self.code

class ADD(Instruction):
  def __init__(self, *operands):
    super(ADD, self).__init__(DuckyOpcodes.ADD, *operands)

class AND(Instruction):
  def __init__(self, *operands):
    super(AND, self).__init__(DuckyOpcodes.AND, *operands)

class BGE(Instruction):
  def __init__(self, label):
    super(BGE, self).__init__(DuckyOpcodes.BRANCH, label)

class BLE(Instruction):
  def __init__(self, label):
    super(BLE, self).__init__(DuckyOpcodes.BRANCH, label)

class BNE(Instruction):
  def __init__(self, label):
    super(BNE, self).__init__(DuckyOpcodes.BRANCH, label)

class BE(Instruction):
  def __init__(self, label):
    super(BE, self).__init__(DuckyOpcodes.BRANCH, label)

class BG(Instruction):
  def __init__(self, label):
    super(BG, self).__init__(DuckyOpcodes.BRANCH, label)

class BL(Instruction):
  def __init__(self, label):
    super(BL, self).__init__(DuckyOpcodes.BRANCH, label)

class CMP(Instruction):
  def __init__(self, left, right):
    super(CMP, self).__init__(DuckyOpcodes.CMP, left, right)

class CALL(Instruction):
  def __init__(self, label):
    super(CALL, self).__init__(DuckyOpcodes.CALL, label)

class HLT(Instruction):
  def __init__(self, isr):
    super(HLT, self).__init__(DuckyOpcodes.HLT, isr)

class INT(Instruction):
  def __init__(self, isr):
    super(INT, self).__init__(DuckyOpcodes.INT, isr)

class INC(Instruction):
  def __init__(self, reg):
    super(INC, self).__init__(DuckyOpcodes.INC, reg)

class J(Instruction):
  def __init__(self, label):
    super(J, self).__init__(DuckyOpcodes.J, label)

class LB(Instruction):
  def __init__(self, reg, addr):
    super(LB, self).__init__(DuckyOpcodes.LB, reg, addr)

class LS(Instruction):
  def __init__(self, reg, addr):
    super(LS, self).__init__(DuckyOpcodes.LS, reg, addr)

class LW(Instruction):
  def __init__(self, reg, addr):
    super(LW, self).__init__(DuckyOpcodes.LW, reg, addr)

class LA(Instruction):
  def __init__(self, reg, value):
    super(LA, self).__init__(DuckyOpcodes.LA, reg, value)

class LI(Instruction):
  def __init__(self, reg, value):
    super(LI, self).__init__(DuckyOpcodes.LI, reg, value)

class MOV(Instruction):
  def __init__(self, *operands):
    super(MOV, self).__init__(DuckyOpcodes.MOV, *operands)

class MUL(Instruction):
  def __init__(self, *operands):
    super(MUL, self).__init__(DuckyOpcodes.MUL, *operands)

class OR(Instruction):
  def __init__(self, *operands):
    super(MUL, self).__init__(DuckyOpcodes.MUL, *operands)

class NOT(Instruction):
  def __init__(self, *operands):
    super(NOT, self).__init__(DuckyOpcodes.NOT, *operands)

class POP(Instruction):
  def __init__(self, reg):
    super(POP, self).__init__(DuckyOpcodes.POP, reg)

class PUSH(Instruction):
  def __init__(self, reg):
    super(PUSH, self).__init__(DuckyOpcodes.PUSH, reg)

class RET(Instruction):
  def __init__(self):
    super(RET, self).__init__(DuckyOpcodes.RET)

class SHL(Instruction):
  def __init__(self, reg, ri):
    super(SHL, self).__init__(DuckyOpcodes.SHIFTL, reg, ri)

class SHR(Instruction):
  def __init__(self, reg, ri):
    super(SHR, self).__init__(DuckyOpcodes.SHIFTR, reg, ri)

class STB(Instruction):
  def __init__(self, addr, reg):
    super(STB, self).__init__(DuckyOpcodes.STB, addr, reg)

class STS(Instruction):
  def __init__(self, addr, reg):
    super(STS, self).__init__(DuckyOpcodes.STS, addr, reg)

class STW(Instruction):
  def __init__(self, addr, reg):
    super(STW, self).__init__(DuckyOpcodes.STW, addr, reg)

class SUB(Instruction):
  def __init__(self, *operands):
    super(SUB, self).__init__(DuckyOpcodes.SUB, *operands)


class NamedValue(object):
  def __init__(self, name):
    super(NamedValue, self).__init__()

    self.name = name

  def __repr__(self):
    return '<%s: %s, rb=%s>' % (self.__class__.__name__, self.name, self.is_register_backed())

  def can_register_backed(self):
    return isinstance(self, StackSlotValue) or isinstance(self, MemorySlotValue)

  def is_register_backed(self):
    return (isinstance(self, StackSlotValue) or isinstance(self, MemorySlotValue)) and self.storage.has_register()

  def backing_register(self):
    assert isinstance(self, StackSlotValue) or isinstance(self, MemorySlotValue)

    return self.storage.register

class RegisterValue(NamedValue):
  def __init__(self, register):
    assert isinstance(register, Register)

    super(RegisterValue, self).__init__(register.name)

    self.register = register

class StackSlotValue(NamedValue):
  def __init__(self, storage):
    super(StackSlotValue, self).__init__('fp[%i]' % storage.offset)

    self.storage = storage

class MemorySlotValue(NamedValue):
  def __init__(self, storage):
    super(MemorySlotValue, self).__init__(storage.label)

    self.storage = storage

class RegisterMemorySlotValue(NamedValue):
  def __init__(self, register):
    super(RegisterMemorySlotValue, self).__init__(register.name)

    self.register = register

class ConstantValue(NamedValue):
  def __init__(self, value):
    super(ConstantValue, self).__init__(value)

    self.value = value

class StringConstantValue(ConstantValue):
  pass

class ExpressionClass(enum.Enum):
  LVALUE  = 0
  MLVALUE = 1
  RVALUE  = 2

class Expression(object):
  def __init__(self, value = None, type = None, klass = ExpressionClass.RVALUE):
    if self.__class__ is Expression:
      raise RuntimeError('It is not possible to create Expression object - use one of its children, [M][LR]ValueExpression instead')

    self.value = value
    self.type = type
    self.klass = klass

  def __repr__(self):
    return '<%s: value=%s, type=%s>' % (self.__class__.__name__, self.value, self.type)

  def is_lvalue(self):
    return self.klass in (ExpressionClass.LVALUE, ExpressionClass.MLVALUE)

  def is_mlvalue(self):
    return self.klass == ExpressionClass.MLVALUE

  def is_rvalue(self):
    return self.klass == ExpressionClass.RVALUE

  def to_rvalue(self, visitor, preferred = None, keep = None):
    assert self.is_lvalue() is True

    visitor.DEBUG(visitor.log_prefix + 'to_rvalue: E=%s', self)
    visitor.DOWN()

    visitor.EMIT(Comment('<lvalue->rvalue conversion: %s' % self))

    RE = RValueExpression(type = self.type.ptr_to_type)

    from .types import ArrayType
    if isinstance(self.type, ArrayType):
      self.WARN(self.log_prefix + 'Implicit array->ptr conversion')

    else:
      if isinstance(self.value, StackSlotValue) or isinstance(self.value, MemorySlotValue):
        storage = self.value.storage

        if storage.has_register():
          visitor.DEBUG(visitor.log_prefix + 'to_rvalue: E is symbol storage already having a register')
          RE.value = RegisterValue(storage.register)

        else:
          visitor.DEBUG(visitor.log_prefix + 'to_rvalue: E is symbol storage without a register')

          RE.value = RegisterValue(visitor.GET_REG(preferred = preferred, keep = keep))
          storage.acquire_register(RE.value.register)

          if len(self.type.ptr_to_type) == 1:
            visitor.EMIT(LB(RE.value.name, self.value.name))

          elif len(self.type.ptr_to_type) == 2:
            visitor.EMIT(LS(RE.value.name, self.value.name))

          elif len(self.type.ptr_to_type) == 4:
            visitor.EMIT(LW(RE.value.name, self.value.name))

          else:
            visitor.WARN(visitor.log_prefix + 'Unhandled lvalue->rvalue size branch')

      else:
        visitor.DEBUG(visitor.log_prefix + 'to_rvalue: E is generic value')

        RE = RValueExpression(value = RegisterValue(visitor.GET_REG(preferred = preferred, keep = keep)), type = self.type.ptr_to_type)

        visitor.DEBUG(visitor.log_prefix + 'RE=%s', RE)

        if len(self.type.ptr_to_type) == 1:
          visitor.EMIT(LB(RE.value.name, self.value.name))

        elif len(self.type.ptr_to_type) == 2:
          visitor.EMIT(LS(RE.value.name, self.value.name))

        elif len(self.type.ptr_to_type) == 4:
          visitor.EMIT(LW(RE.value.name, self.value.name))

        else:
          visitor.WARN(visitor.log_prefix + 'Unhandled lvalue->rvalue size branch')

    visitor.DEBUG(visitor.log_prefix + 'RE=%s', RE)

    visitor.EMIT(Comment('</ lvalue->rvalue conversion: %s>' % RE))
    visitor.UP()
    return RE, self

class LValueExpression(Expression):
  def __init__(self, *args, **kwargs):
    super(LValueExpression, self).__init__(klass = ExpressionClass.LVALUE, *args, **kwargs)

class MLValueExpression(Expression):
  def __init__(self, *args, **kwargs):
    super(MLValueExpression, self).__init__(klass = ExpressionClass.MLVALUE, *args, **kwargs)

class RValueExpression(Expression):
  def __init__(self, *args, **kwargs):
    super(RValueExpression, self).__init__(klass = ExpressionClass.RVALUE, *args, **kwargs)

class Block(object):
  id = 0

  def __init__(self, name = None, comment = None):
    Block.id += 1
    self.id = Block.id

    self.names = [name] if name is not None else []

    self.comment = comment

    self.code = []

    self.incoming = {}
    self.outgoing = {}

  def __repr__(self):
    return 'Block(id=%i, names=%s, in=%s, out=%s%s)' % (self.id, ','.join(self.names), ','.join([str(block.id) for block in itervalues(self.incoming)]), ','.join([str(block.id) for block in itervalues(self.outgoing)]), (' (%s)' % self.comment) if self.comment is not None else '')

  def instructions(self):
    return [i for i in self.code if isinstance(i, Instruction)]

  def add_name(self, name):
    self.names.append(name)

  def emit(self, inst):
    self.code.append(inst)

  def add_outgoing(self, block):
    self.outgoing[block.id] = block

  def add_incoming(self, block):
    self.incoming[block.id] = block

  def connect(self, next):
    self.outgoing[next.id] = next
    next.incoming[self.id] = self

  def materialize(self, code):
    code.append('')
    code.append('  ; block: id=%s, names=%s, in=%s, out=%s%s' % (self.id, ', '.join(self.names), ', '.join([str(block.id) for block in itervalues(self.incoming)]), ', '.join([str(block.id) for block in itervalues(self.outgoing)]), (' (%s)' % self.comment) if self.comment is not None else ''))

    for name in self.names:
      code.append('%s:' % name)

    for inst in self.code:
      if hasattr(inst, 'materialize'):
        code.append('  ' + inst.materialize())

      else:
        code.append(inst)

class Function(object):
  def __init__(self, visitor, decl, ftype, args_types = None):
    self.visitor = visitor
    self.decl = decl
    self.registers = RegisterSet(self)

    self.name = decl.name

    self.type = ftype

    self.fp_offset = 0

    self.blocks = []

    self._header_block = self.block(comment = '%s header' % decl.name)
    self._prolog_block = self.block(name = decl.name, comment = '%s prolog' % decl.name)
    self._args_block   = self.block(comment = '%s args' % decl.name)
    self._body_block   = self.block(comment = '%s body' % decl.name)
    self._epilog_block = Block(name = visitor.get_new_label(name = self.name + '_return'), comment = '%s epilog' % decl.name)

    self._header_block.connect(self._prolog_block)
    self._prolog_block.connect(self._args_block)
    self._args_block.connect(self._body_block)

  def block(self, *args, **kwargs):
    block = Block(*args, **kwargs)
    self.blocks.append(block)
    return block

  def header_block(self):
    return self._header_block

  def prolog_block(self):
    return self._prolog_block

  def args_block(self):
    return self._args_block

  def body_block(self):
    return self._body_block

  def epilog_block(self):
    return self._epilog_block

  def finish(self):
    self.header_block().emit(Directive('.text'))

    if 'static' not in self.decl.storage:
      self.header_block().emit(Directive('.global {name}'.format(name = self.decl.name)))

    self.blocks[-1].connect(self.epilog_block())
    self.blocks.append(self.epilog_block())

    self.prolog_block().emit(SUB('sp', abs(self.fp_offset)))
    self.registers.save_callee_saves(self.prolog_block())

    self.registers.restore_callee_saves(self.epilog_block())
    self.epilog_block().emit(ADD('sp', abs(self.fp_offset)))
    self.epilog_block().emit(RET())

  def materialize(self):
    code = []

    for block in self.blocks:
      block.materialize(code)

    return code



import sys
import argparse

from six import iterkeys

from . import parse_options
from ..cc import CompilerError
from ..cc.passes import AST_PASSES, BT_PASSES

from itertools import chain
from pycparser import parse_file

def compile_file(logger, options, file_in, file_out):
  ast = parse_file(file_in, use_cpp = True, cpp_path = 'gcc', cpp_args = ['-E', r'-Ilibc/include'])

  # ast.show()

  options.passes = list(set(options.passes))

  def __apply_opt_passes(passes, tree):
    tree_modified = True
    while tree_modified is True:
      tree_modified = False

      for opt_pass in iterkeys(passes):
        if opt_pass not in options.passes or ('no-' + opt_pass) in options.passes:
          logger.debug('Pass %s is disabled', opt_pass)
          continue

        logger.debug('Pass %s is enabled, running', opt_pass)

        v = passes[opt_pass](logger)
        v.DOWN()

        v.tree_modified = True
        while v.tree_modified is True:
          v.tree_modified = False
          v.visit(tree)

          if v.tree_modified is True:
            tree_modified = True

        v.UP()

  __apply_opt_passes(AST_PASSES, ast)

  cv = AST_PASSES['ast-codegen'](logger)

  try:
    cv.visit(ast)

  except Exception:
    logger.exception('Exception raised during compilation')
    raise SystemExit(1)

  __apply_opt_passes(BT_PASSES, cv)

  with open(file_out, 'w') as f_out:
    f_out.write(cv.materialize())

def main():
  parser = argparse.ArgumentParser()

  group = parser.add_argument_group('Tool verbosity')
  group.add_argument('-d', '--debug', dest = 'debug', action = 'store_true', default = False, help = 'Debug mode')
  group.add_argument('-q', '--quiet', dest = 'quiet', action = 'count',      default = 0,     help = 'Decrease verbosity. This option can be used multiple times')
  group.add_argument('-v', '--verbose', dest = 'verbose', action = 'count',  default = 0,     help = 'Increase verbosity. This option can be used multiple times')

  group = parser.add_argument_group('File options')
  group.add_argument('-i', dest = 'file_in',  action = 'append',     default = [],    help = 'Input file')
  group.add_argument('-o', dest = 'file_out', action = 'append',     default = [],    help = 'Output file')
  group.add_argument('-f', dest = 'force',    action = 'store_true', default = False, help = 'Force overwrite of the output file')

  group = parser.add_argument_group('Passes')
  group.add_argument('-O', dest = 'opt_level', action = 'store',      default = 1,     nargs = 1, type = int, help = 'Set requested optimization level', metavar = 'LEVEL')
  group.add_argument('-p', dest = 'passes',    action = 'append',     default = [],    nargs = 1, help = 'Enable or disable optimization pass', metavar = 'PASS')

  options, logger = parse_options(parser)

  from ..cc.passes import load as load_passes
  load_passes(logger)

  options.passes = list(chain.from_iterable(options.passes))

  if options.opt_level == [0]:
    for p in iterkeys(AST_PASSES):
      options.passes.append('no-%s' % p)

    for p in iterkeys(BT_PASSES):
      options.passes.append('no-%s' % p)

  elif options.opt_level == [1]:
    options.passes += ['ast-const-prop', 'ast-dce', 'bt-peephole', 'bt-simplify']

  else:
    pass

  if not options.file_in:
    parser.print_help()
    sys.exit(1)

  try:
    for filepath in options.file_in:
      compile_file(logger, options, filepath, options.file_out.pop(0))

  except CompilerError as e:
    logger.error(e.message)



"""
Coprocessors are intended to extend operations of CPUs. They are optional, and
can cover wide range of operations, e.g. floating point arithmetic, encryption,
or graphics. They are always attached to a CPU core, and may contain and use
internal resources, e.g. their very own register sets, machine's memory, or
their parent's caches.
"""

class Coprocessor(object):
  """
  Base class for per-core coprocessors.
  """

  def __init__(self, core):
    super(Coprocessor, self).__init__()

    self.core = core



import collections
import functools
import string

from six import iteritems, integer_types, PY2

from ctypes import sizeof

def align(boundary, n):
  return (n + boundary - 1) & ~(boundary - 1)

def str2int(s):
  if isinstance(s, integer_types):
    return s

  if s.startswith('0x'):
    return int(s, base = 16)

  if s.startswith('0'):
    return int(s, base = 8)

  return int(s)

def sizeof_fmt(n, suffix = 'B', max_unit = 'Zi'):
  for unit in ['', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi']:
    if abs(n) < 1024.0 or max_unit == unit:
      return "%3.1f%s%s" % (n, unit, suffix)

    n /= 1024.0

  return "%.1f%s%s" % (n, 'Yi', suffix)

class Formatter(string.Formatter):
  def format_field(self, value, format_spec):
    if format_spec and format_spec[-1] in 'BSWL':
      return self.format_int(format_spec, value)

    return super(Formatter, self).format_field(value, format_spec)

  def format_int(self, format_spec, value):
    i = value if isinstance(value, integer_types) else value.value

    if format_spec.endswith('B'):
      return '0x{:02X}'.format(i & 0xFF)

    if format_spec.endswith('S'):
      return '0x{:04X}'.format(i & 0xFFFF)

    if format_spec.endswith('W'):
      return '0x{:08X}'.format(i & 0xFFFFFFFF)

    if format_spec.endswith('L'):
      return '0x{:016X}'.format(i & 0xFFFFFFFFFFFFFFFF)

    return '{:d}'.format(i)

_F = Formatter()
F = _F.format

UINT8_FMT  = functools.partial(_F.format_int, 'B')
UINT16_FMT = functools.partial(_F.format_int, 'S')
UINT32_FMT = functools.partial(_F.format_int, 'W')
UINT64_FMT = functools.partial(_F.format_int, 'L')


if PY2:
  _BaseFile = file  # noqa

  bytes2str = str
  str2bytes = str
  int2bytes = chr

else:
  from io import IOBase as _BaseFile
  import functools

  bytes2str = functools.partial(str, encoding = 'latin-1')
  str2bytes = functools.partial(bytes, encoding = 'latin-1')

  def int2bytes(b):
    return bytes([b])

def isfile(o):
  return isinstance(o, (_BaseFile, BinaryFile))


class BinaryFile(object):
  """
  Base class of all classes that represent "binary" files - binaries, core dumps.
  It provides basic methods for reading and writing structures.
  """

  @staticmethod
  def do_open(logger, path, mode = 'rb', klass = None):
    if 'b' not in mode:
      mode += 'b'

    stream = open(path, mode)

    if not PY2:
      import io

      if 'r' in mode:
        if 'w' in mode:
          stream = io.BufferedRandom(stream)

        else:
          stream = io.BufferedReader(stream)

      else:
        stream = io.BufferedWriter(stream)

    klass = klass or BinaryFile

    return klass(logger, stream)

  @staticmethod
  def open(*args, **kwargs):
    return BinaryFile.do_open(*args, **kwargs)

  def __init__(self, logger, stream):
    self.stream = stream

    self.DEBUG = logger.debug
    self.INFO = logger.info
    self.WARN = logger.warning
    self.ERROR = logger.error
    self.EXCEPTION = logger.exception

    self.close = stream.close
    self.flush = stream.flush
    self.name = stream.name
    self.read = stream.read
    self.readinto = stream.readinto
    self.readline = stream.readline
    self.seek = stream.seek
    self.tell = stream.tell
    self.write = stream.write

    self.setup()

  def __enter__(self):
    return self

  def __exit__(self, *args, **kwargs):
    self.close()

  def setup(self):
    pass

  def read_struct(self, st_class):
    """
    Read structure from current position in file.

    :returns: instance of class ``st_class`` with content read from file
    :rtype: ``st_class``
    """

    pos = self.tell()

    st = st_class()
    self.readinto(st)

    self.DEBUG('read_struct: %s: %s bytes: %s', pos, sizeof(st_class), st)

    return st

  def write_struct(self, st):
    """
    Write structure into file at the current position.

    :param class st: ``ctype``-based structure
    """

    pos = self.tell()

    self.DEBUG('write_struct: %s: %s bytes: %s', pos, sizeof(st), st)

    self.write(st)

class LRUCache(collections.OrderedDict):
  """
  Simple LRU cache, based on ``OrderedDict``, with limited size. When limit
  is reached, the least recently inserted item is removed.

  :param logging.Logger logger: logger object instance should use for logging.
  :param int size: maximal number of entries allowed.
  """

  def __init__(self, logger, size, *args, **kwargs):
    super(LRUCache, self).__init__(*args, **kwargs)

    self.size = size

    self.reads   = 0
    self.inserts = 0
    self.hits    = 0
    self.misses  = 0
    self.prunes  = 0

    self.DEBUG = logger.debug
    self.INFO = logger.info
    self.WARN = logger.warn
    self.ERROR = logger.error
    self.EXCEPTION = logger.exception

  def make_space(self):
    """
    This method is called when there is no free space in cache. It's responsible
    for freeing at least one slot, upper limit of removed entries is not enforced.
    """

    self.DEBUG('%s.make_space', self.__class__.__name__)

    self.popitem(last = False)
    self.prunes += 1

  def __getitem__(self, key):
    """
    Return entry with specified key.
    """

    self.DEBUG('%s.__getitem__: key=%s', self.__class__.__name__, key)

    self.reads += 1

    if key in self:
      self.hits += 1
    else:
      self.misses += 1

    return super(LRUCache, self).__getitem__(key)

  def __setitem__(self, key, value):
    """
    Called when item is inserted into cache. Size limit is checked and if there's no free
    space in cache, ``make_space`` method is called.
    """

    self.DEBUG('%s.__setitem__: key=%s, value=%s', self.__class__.__name__, key, value)

    if len(self) == self.size:
      self.make_space()

    super(LRUCache, self).__setitem__(key, value)
    self.inserts += 1

  def get_object(self, key):
    """
    The real workhorse - responsible for getting requested item from outside when it's
    not present in cache. Called by ``__missing__`` method. This method itself makes no
    changes to cache at all.
    """

    raise NotImplementedError('Cache %s does not implement get() method' % self.__class__.__name__)

  def __missing__(self, key):
    """
    Called when requested entry is not in cache. It's responsible for getting missing item
    and inserting it into cache. Returns new item.
    """

    self.DEBUG('%s.__missing__: key=%s', self.__class__.__name__, key)

    self[key] = value = self.get_object(key)
    return value

class StringTable(object):
  """
  Simple string table, used by many classes operating with files (core, binaries, ...).
  String can be inserted into table and read, each has its starting offset and its end is
  marked with null byte (\0).

  Thsi is a helper class - it makes working with string, e.g. section and symbol names,
  much easier.
  """

  def __init__(self):
    super(StringTable, self).__init__()

    self.buff = ''

  def put_string(self, s):
    """
    Insert new string into table. String is appended at the end of internal buffer,
    and terminating zero byte (\0) is appended to mark end of string.

    :returns: offset of inserted string
    :rtype: ``int``
    """

    offset = len(self.buff)

    self.buff += chr(len(s))
    self.buff += s

    return offset

  def get_string(self, offset):
    """
    Read string from table.

    :param int offset: offset of the first character from the beginning of the table
    :returns: string
    :rtype: ``string``
    """

    l = ord(self.buff[offset])
    return ''.join(self.buff[offset + 1:offset + 1 + l])

class SymbolTable(dict):
  def __init__(self, binary):
    self.binary = binary

  def __getitem__(self, address):
    last_symbol = None
    last_symbol_offset = 0xFFFE

    for symbol_name, symbol in iteritems(self.binary.symbols):
      if symbol.address > address:
        continue

      if symbol.address == address:
        return (symbol_name, 0)

      offset = abs(address - symbol.address)
      if offset < last_symbol_offset:
        last_symbol = symbol_name
        last_symbol_offset = offset

    return (last_symbol, last_symbol_offset)

  def get_symbol(self, name):
    return self.binary.symbols[name]


class Flags(object):
  _flags = []
  _labels = ''
  _encoding = []  # silence Codacy warning - _encoding will have a real value

  @classmethod
  def create(cls, **kwargs):
    flags = cls()

    for name in cls._flags:
      setattr(flags, name, True if kwargs.get(name, False) is True else False)

    return flags

  @classmethod
  def encoding(cls):
    return cls._encoding

  @classmethod
  def from_encoding(cls, encoding):
    flags = cls()
    flags.load_encoding(encoding)
    return flags

  def to_encoding(self):
    encoding = self._encoding()
    self.save_encoding(encoding)
    return encoding

  def load_encoding(self, encoding):
    for name in [field[0] for field in encoding._fields_]:
      setattr(self, name, True if getattr(encoding, name) == 1 else False)

  def save_encoding(self, encoding):
    for name in [field[0] for field in encoding._fields_]:
      setattr(encoding, name, 1 if getattr(self, name) is True else 0)

  def to_int(self):
    u = 0

    for i, name in enumerate(self._flags):
      if getattr(self, name) is True:
        u |= (1 << i)

    return u

  def load_int(self, u):
    for i, name in enumerate(self._flags):
      setattr(self, name, True if u & (1 << i) else False)

  @classmethod
  def from_int(cls, u):
    flags = cls()
    flags.load_int(u)
    return flags

  def to_string(self):
    return ''.join([
      self._labels[i] if getattr(self, name) is True else '-' for i, name in enumerate(self._flags)
    ])

  def load_string(self, s):
    s = s.upper()

    for i, name in enumerate(self._flags):
      setattr(self, name, True if self._labels[i] in s else False)

  @classmethod
  def from_string(cls, s):
    flags = cls()
    flags.load_string(s)
    return flags

  def __repr__(self):
    return '<{}: {}>'.format(self.__class__.__name__, self.to_string())



try:
  import cPickle as pickle

except ImportError:
  import pickle

import collections
import optparse
import sys

from six import itervalues

def read_profiling_data(logger, files_in):
  from ..profiler import ProfileRecord

  data = collections.defaultdict(ProfileRecord)

  for path in files_in:
    logger.info('Reading profile data from %s', path)

    with open(path, 'rb') as f_in:
      file_data = pickle.load(f_in)

    for record in itervalues(file_data):
      data[record.ip].merge(record)

    logger.debug('%d records merged', len(file_data))

  return data

def main():
  from . import add_common_options, parse_options

  parser = optparse.OptionParser()
  add_common_options(parser)

  parser.add_option('-i', dest = 'files_in', action = 'append', default = [], help = 'Input file')
  parser.add_option('-b', dest = 'binary', action = 'store', default = None, help = 'Profiled binary')

  options, logger = parse_options(parser)

  if not options.files_in:
    parser.print_help()
    sys.exit(1)

  if not options.binary:
    parser.print_help()
    sys.exit(1)

  merged_data = read_profiling_data(logger, options.files_in)

  from ..mm.binary import File
  from ..util import SymbolTable

  binary = File.open(logger, options.binary, 'r')
  binary.load()
  binary.load_symbols()

  symbol_table = SymbolTable(binary)

  all_hits = sum([record.count for record in itervalues(merged_data)])

  def print_points():
    from ..mm import UINT32_FMT
    from ..cpu.instructions import DuckyInstructionSet

    table = [
      ['Address', 'Symbol', 'Offset', 'Hits', 'Percentage', 'Inst']
    ]

    for ip in sorted(merged_data.keys(), key = lambda x: merged_data[x].count, reverse = True):
      record = merged_data[ip]
      #binary_ip = record.ip - DEFAULT_BOOTLOADER_ADDRESS
      binary_ip  = record.ip

      symbol, offset = symbol_table[binary_ip]

      if symbol is None:
        symbol_name = ''
        inst_disassembly = '<unknown>'

      else:
        symbol_name = symbol

        symbol = symbol_table.get_symbol(symbol)
        header, content = binary.get_section(symbol.section)

        inst_encoding, inst_cls, inst_opcode = DuckyInstructionSet.decode_instruction(logger, content[(symbol.address + offset - header.base) / 4].value)
        inst_disassembly = DuckyInstructionSet.disassemble_instruction(logger, inst_encoding)

      table.append([UINT32_FMT(binary_ip), symbol_name, UINT32_FMT(offset), record.count, '%.02f' % (float(record.count) / float(all_hits) * 100.0), inst_disassembly])

    logger.table(table)

  print_points()



#! /usr/bin/env python

import sys

from mako.template import Template
from functools import partial

def parse_template(file_in, file_out):
  def _X(i, padding = None):
    padding = ('0' + str(padding)) if padding is not None else ''

    return ('0x%' + padding + 'X') % i

  X = _X
  X2 = partial(_X, padding = 2)
  X4 = partial(_X, padding = 4)
  X8 = partial(_X, padding = 8)

  with open(file_in, 'r') as f_in:
    with open(file_out, 'w') as f_out:
      f_out.write(Template(f_in.read()).render(X = X, X2 = X2, X4 = X4, X8 = X8))

def main():
  import optparse
  from . import add_common_options, parse_options

  parser = optparse.OptionParser()
  add_common_options(parser)

  group = optparse.OptionGroup(parser, 'File options')
  parser.add_option_group(group)
  group.add_option('-i', dest = 'file_in', default = None, help = 'Input file')
  group.add_option('-o', dest = 'file_out', default = None, help = 'Output file')
  group.add_option('-f', dest = 'force', default = False, action = 'store_true', help = 'Force overwrite of the output file')

  options, logger = parse_options(parser)

  if not options.file_in:
    parser.print_help()
    sys.exit(1)

  parse_template(options.file_in, options.file_out)



import ctypes
import enum
import re
import sys

from six import integer_types, string_types, add_metaclass
from six.moves import range
from functools import partial

from .registers import Registers, REGISTER_NAMES
from ..mm import u32_t, i32_t, UINT16_FMT, UINT32_FMT
from ..util import str2int
from ..errors import EncodingLargeValueError, UnalignedJumpTargetError, AccessViolationError, AssemblerError

PO_REGISTER  = r'(?P<register_n{operand_index}>(?:r\d\d?)|(?:sp)|(?:fp))'
PO_AREGISTER = r'(?P<address_register>r\d\d?|sp|fp)(?:\[(?:(?P<offset_sign>-|\+)?(?P<offset_immediate>0x[0-9a-fA-F]+|\d+))\])?'
PO_IMMEDIATE = r'(?:(?P<immediate>(?:-|\+)?(?:0x[0-9a-fA-F]+|\d+))|(?P<immediate_address>&[a-zA-Z_\.][a-zA-Z0-9_]*))'

def UINT20_FMT(i):
  return '0x%05X' % (i & 0xFFFFF)

def encoding_to_u32(inst):
  return ctypes.cast(ctypes.byref(inst), ctypes.POINTER(u32_t)).contents.value

if hasattr(sys, 'pypy_version_info'):
  def u32_to_encoding(u, encoding):
    class _Cast(ctypes.Union):
      _pack_ = 0
      _fields_ = [
        ('overall',  u32_t),
        ('encoding', encoding)
      ]

    caster = _Cast()
    caster.overall = u32_t(u).value
    return caster.encoding

else:
  def u32_to_encoding(u, encoding):
    u = u32_t(u)
    e = encoding()

    ctypes.cast(ctypes.byref(e), ctypes.POINTER(encoding))[0] = ctypes.cast(ctypes.byref(u), ctypes.POINTER(encoding)).contents

    return e

def IE_OPCODE():
  return ('opcode', u32_t, 6)

def IE_FLAG(n):
  return (n, u32_t, 1)

def IE_REG(n):
  return (n, u32_t, 5)

def IE_IMM(n, l):
  return (n, u32_t, l)

class Encoding(ctypes.LittleEndianStructure):
  @staticmethod
  def sign_extend_immediate(logger, inst, sign_mask, ext_mask):
    logger.debug('sign_extend_immediate: inst=%s, sign_mask=%s, ext_mask=%s', inst, UINT32_FMT(sign_mask), UINT32_FMT(ext_mask))

    if __debug__:
      u = u32_t(ext_mask | inst.immediate) if inst.immediate & sign_mask else u32_t(inst.immediate)
      logger.debug('  result=%s', UINT32_FMT(u))

      return u.value

    else:
      i = inst.immediate
      return ((ext_mask | i) & 0xFFFFFFFF) if i & sign_mask else i
      return u32_t(ext_mask | inst.immediate).value if inst.immediate & sign_mask else u32_t(inst.immediate).value

class EncodingR(ctypes.LittleEndianStructure):
  _pack_ = 0
  _fields_ = [
    IE_OPCODE(),                # 0
    IE_REG('reg1'),             # 6
    IE_REG('reg2'),             # 11
    IE_FLAG('immediate_flag'),  # 16
    IE_IMM('immediate', 15),    # 17
  ]

  @staticmethod
  def fill_reloc_slot(logger, inst, slot):
    logger.debug('fill_reloc_slot: inst=%s, slot=%s', inst, slot)

    slot.patch_offset = 17
    slot.patch_size = 15

  @staticmethod
  def sign_extend_immediate(logger, inst):
    return Encoding.sign_extend_immediate(logger, inst, 0x4000, 0xFFFF8000)

  def __repr__(self):
    return '<EncodingR: opcode=%s, reg1=%s, reg2=%s, immediate_flag=%s, immediate=%s>' % (self.opcode, self.reg1, self.reg2, self.immediate_flag, UINT16_FMT(self.immediate))

class EncodingC(ctypes.LittleEndianStructure):
  _pack_ = 0
  _fields_ = [
    IE_OPCODE(),                # 0
    IE_REG('reg'),              # 6
    IE_IMM('flag', 3),          # 11
    IE_FLAG('value'),           # 14
    IE_FLAG('immediate_flag'),  # 15
    IE_IMM('immediate', 16)     # 16
  ]

  @staticmethod
  def fill_reloc_slot(logger, inst, slot):
    logger.debug('fill_reloc_slot: inst=%s, slot=%s', inst, slot)

    slot.patch_offset = 16
    slot.patch_size = 16

  @staticmethod
  def sign_extend_immediate(logger, inst):
    return Encoding.sign_extend_immediate(logger, inst, 0x8000, 0xFFFF0000)

  def __repr__(self):
    return '<EncodingC: opcode=%s, reg=%s, flag=%s, value=%s, immediate_flag=%s, immediate=%s>' % (self.opcode, self.reg, self.flag, self.value, self.immediate_flag, UINT16_FMT(self.immediate))

class EncodingI(ctypes.LittleEndianStructure):
  _pack_ = 0
  _fields_ = [
    IE_OPCODE(),                # 0
    IE_REG('reg'),              # 6
    IE_FLAG('immediate_flag'),  # 11
    IE_IMM('immediate', 20),    # 12
  ]

  @staticmethod
  def fill_reloc_slot(logger, inst, slot):
    logger.debug('fill_reloc_slot: inst=%s, slot=%s', inst, slot)

    slot.patch_offset = 12
    slot.patch_size = 20

  @staticmethod
  def sign_extend_immediate(logger, inst):
    return Encoding.sign_extend_immediate(logger, inst, 0x80000, 0xFFF00000)

  def __repr__(self):
    return '<EncodingI: opcode=%s, reg=%s, immediate_flag=%s, immediate=%s>' % (self.opcode, self.reg, self.immediate_flag, UINT20_FMT(self.immediate))

class EncodingA(ctypes.LittleEndianStructure):
  _pack_ = 0
  _fields_ = [
    IE_OPCODE(),                # 0
    IE_REG('reg1'),             # 6
    IE_REG('reg2'),             # 11
    IE_REG('reg3')              # 16
  ]

  def __repr__(self):
    return '<EncodingA: opcode=%s, reg1=%s, reg2=%s, reg3=%s>' % (self.opcode, self.reg1, self.reg2, self.reg3)

def ENCODE(logger, buffer, inst, field, size, value, raise_on_large_value = False):
  logger.debug('ENCODE: inst=%s, field=%s, size=%s, value=%s, raise_on_large_value=%s', inst, field, size, value, raise_on_large_value)

  setattr(inst, field, value)

  logger.debug('ENCODE: inst=%s', inst)

  if value >= 2 ** size:
    e = buffer.get_error(EncodingLargeValueError, 'inst=%s, field=%s, size=%s, value=%s' % (inst, field, size, UINT32_FMT(value)))

    if raise_on_large_value is True:
      raise e

    e.log(logger.warn)

class Descriptor(object):
  mnemonic      = None
  opcode        = None
  operands      = None

  # this is a default encoding, and by the way it silents Codacy's warning
  encoding      = EncodingR

  pattern       = None

  relative_address = False
  inst_aligned = False

  def __init__(self, instruction_set):
    super(Descriptor, self).__init__()

    self.instruction_set = instruction_set

    pattern = r'\s*' + self.mnemonic

    if self.operands:
      operand_patterns = []

      self.operands = [ot.strip() for ot in self.operands.split(',')]

      for operand_index, operand_types in zip(list(range(0, len(self.operands))), self.operands):
        operand_pattern = []

        for operand_type in operand_types:
          if operand_type == 'r':
            operand_pattern.append(PO_REGISTER.format(operand_index = operand_index))

          elif operand_type == 'a':
            operand_pattern.append(PO_AREGISTER)

          elif operand_type == 'i':
            operand_pattern.append(PO_IMMEDIATE)

          else:
            raise Exception('Unhandled operand type: {}'.format(operand_type))

        operand_patterns.append('(?:' + '|'.join(operand_pattern) + ')')

      pattern += r' ' + r',\s*'.join(operand_patterns)

    pattern = r'^' + pattern + '(?:\s*[;#].*)?$'
    self.pattern = re.compile(pattern, re.MULTILINE)

    self.instruction_set.instructions.append(self)

  @staticmethod
  def jit(core, inst):
    return None

  @staticmethod
  def execute(core, inst):
    raise NotImplementedError('%s does not implement execute method' % inst.opcode)

  @staticmethod
  def assemble_operands(logger, buffer, inst, operands):
    pass

  @staticmethod
  def fill_reloc_slot(logger, inst, slot):
    inst.fill_reloc_slot(logger, inst, slot)

  @staticmethod
  def disassemble_operands(logger, inst):
    return []

  @classmethod
  def disassemble_mnemonic(cls, inst):
    return cls.mnemonic

  def emit_instruction(self, logger, buffer, line):
    DEBUG = logger.debug

    DEBUG('emit_instruction: input line: %s', line)

    binst = self.encoding()

    DEBUG('emit_instruction: encoding=%s', self.encoding.__name__)
    DEBUG('emit_instruction: desc is %s', self)

    binst.opcode = self.opcode

    raw_match = self.pattern.match(line)
    matches = raw_match.groupdict()
    DEBUG('emit_instruction: matches=%s', matches)

    operands = {}

    def str2reg(r):
      if r == 'sp':
        return Registers.SP

      if r == 'fp':
        return Registers.FP

      return Registers(int(r[1:]))

    if self.operands and len(self.operands):
      for operand_index in range(0, len(self.operands)):
        reg_group_name = 'register_n{}'.format(operand_index)

        if reg_group_name in matches and matches[reg_group_name]:
          operands[reg_group_name] = str2reg(matches[reg_group_name])

        elif 'address_register' in matches and matches['address_register']:
          operands['areg'] = str2reg(matches['address_register'])

          if 'pointer' in matches and matches['pointer'] is not None:
            if matches['pointer'] == '[':
              operands['pointer'] = 'offset'

            elif matches['pointer'] == '(':
              operands['pointer'] = 'segment'

            else:
              raise Exception('Unhandled pointer type: {}'.format(matches))

          if 'offset_register' in matches and matches['offset_register'] is not None:
            operands['offset_register'] = str2reg(matches['offset_register'])

          elif 'offset_immediate' in matches and matches['offset_immediate'] is not None:
            k = -1 if 'offset_sign' in matches and matches['offset_sign'] and matches['offset_sign'].strip() == '-' else 1
            operands['offset_immediate'] = k * str2int(matches['offset_immediate'])

        elif 'immediate' in matches and matches['immediate']:
          operands['immediate'] = str2int(matches['immediate'])

        elif 'immediate_address' in matches and matches['immediate_address'] is not None:
          operands['immediate'] = matches['immediate_address']

        else:
          raise Exception('Unhandled operand: {}'.format(matches))

    else:
      pass

    try:
      self.assemble_operands(logger, buffer, binst, operands)

    except AssemblerError as e:
      e.location = buffer.location.copy()
      raise e

    return binst

class Descriptor_R(Descriptor):
  operands = 'r'
  encoding = EncodingR

  @staticmethod
  def assemble_operands(logger, buffer, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    ENCODE(logger, buffer, inst, 'reg1', 5, operands['register_n0'])

  @staticmethod
  def disassemble_operands(logger, inst):
    return [REGISTER_NAMES[inst.reg1]]

class Descriptor_I(Descriptor):
  operands = 'i'

  @staticmethod
  def assemble_operands(logger, buffer, inst, operands):
    from .assemble import Reference

    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    v = operands['immediate']

    if isinstance(v, integer_types):
      inst.immediate = v

    elif isinstance(v, string_types):
      inst.refers_to = Reference(label = v)

  @staticmethod
  def disassemble_operands(logger, inst):
    return [str(inst.refers_to) if hasattr(inst, 'refers_to') and inst.refers_to is not None else UINT16_FMT(inst.immediate)]

class Descriptor_RI(Descriptor):
  operands = 'ri'
  encoding = EncodingI

  @staticmethod
  def assemble_operands(logger, buffer, inst, operands):
    from .assemble import Reference

    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    if 'register_n0' in operands:
      ENCODE(logger, buffer, inst, 'reg', 5, operands['register_n0'])

    else:
      v = operands['immediate']

      ENCODE(logger, buffer, inst, 'immediate_flag', 1, 1)

      if isinstance(v, integer_types):
        ENCODE(logger, buffer, inst, 'immediate', 20, v)

      elif isinstance(v, string_types):
        inst.refers_to = Reference(label = v)

  @staticmethod
  def disassemble_operands(logger, inst):
    if inst.immediate_flag == 0:
      return [REGISTER_NAMES[inst.reg]]

    return [str(inst.refers_to) if hasattr(inst, 'refers_to') and inst.refers_to is not None else UINT32_FMT(inst.immediate)]

class Descriptor_R_I(Descriptor):
  operands = 'r,i'
  encoding = EncodingI

  @staticmethod
  def assemble_operands(logger, buffer, inst, operands):
    from .assemble import Reference

    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    ENCODE(logger, buffer, inst, 'reg', 5, operands['register_n0'])
    ENCODE(logger, buffer, inst, 'immediate_flag', 1, 1)

    v = operands['immediate']

    if isinstance(v, integer_types):
      ENCODE(logger, buffer, inst, 'immediate', 20, v)

    elif isinstance(v, string_types):
      inst.refers_to = Reference(label = v)

  @staticmethod
  def disassemble_operands(logger, inst):
    return [REGISTER_NAMES[inst.reg], str(inst.refers_to) if hasattr(inst, 'refers_to') and inst.refers_to is not None else UINT20_FMT(inst.immediate)]

class Descriptor_R_RI(Descriptor):
  operands = 'r,ri'
  encoding = EncodingR

  @staticmethod
  def assemble_operands(logger, buffer, inst, operands):
    from .assemble import Reference

    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    ENCODE(logger, buffer, inst, 'reg1', 5, operands['register_n0'])

    if 'register_n1' in operands:
      ENCODE(logger, buffer, inst, 'reg2', 5, operands['register_n1'])

    else:
      ENCODE(logger, buffer, inst, 'immediate_flag', 1, 1)

      v = operands['immediate']

      if isinstance(v, integer_types):
        ENCODE(logger, buffer, inst, 'immediate', 15, v)

      elif isinstance(v, string_types):
        inst.refers_to = Reference(label = v)

  @staticmethod
  def disassemble_operands(logger, inst):
    if inst.immediate_flag == 0:
      return [REGISTER_NAMES[inst.reg1], REGISTER_NAMES[inst.reg2]]

    return [REGISTER_NAMES[inst.reg1], str(inst.refers_to) if hasattr(inst, 'refers_to') and inst.refers_to is not None else UINT16_FMT(inst.immediate)]

class Descriptor_R_R(Descriptor):
  operands = 'r,r'
  encoding = EncodingR

  @staticmethod
  def assemble_operands(logger, buffer, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    ENCODE(logger, buffer, inst, 'reg1', 5, operands['register_n0'])
    ENCODE(logger, buffer, inst, 'reg2', 5, operands['register_n1'])

  @staticmethod
  def disassemble_operands(logger, inst):
    return [REGISTER_NAMES[inst.reg1], REGISTER_NAMES[inst.reg2]]

class Descriptor_RI_R(Descriptor):
  operands = 'ri,r'
  encoding = EncodingR

  @staticmethod
  def assemble_operands(logger, buffer, inst, operands):
    from .assemble import Reference

    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    ENCODE(logger, buffer, inst, 'reg2', 5, operands['register_n1'])

    if 'register_n0' in operands:
      ENCODE(logger, buffer, inst, 'reg1', 5, operands['register_n0'])

    else:
      ENCODE(logger, buffer, inst, 'immediate_flag', 1, 1)

      v = operands['immediate']

      if isinstance(v, integer_types):
        ENCODE(logger, buffer, inst, 'immediate', 15, v)

      elif isinstance(v, string_types):
        inst.refers_to = Reference(label = v)

  @staticmethod
  def disassemble_operands(logger, inst):
    if inst.immediate_flag == 0:
      return [REGISTER_NAMES[inst.reg1], REGISTER_NAMES[inst.reg2]]

    return [str(inst.refers_to) if hasattr(inst, 'refers_to') and inst.refers_to is not None else UINT16_FMT(inst.immediate), REGISTER_NAMES[inst.reg2]]

class InstructionSetMetaclass(type):
  def __init__(cls, name, bases, dict):
    cls.instructions = []

@add_metaclass(InstructionSetMetaclass)
class InstructionSet(object):
  instruction_set_id = None
  opcodes = None

  @classmethod
  def init(cls):
    if hasattr(cls, 'opcode_desc_map'):
      return

    cls.opcode_desc_map = {}
    cls.opcode_encoding_map = {}

    for desc in cls.instructions:
      cls.opcode_desc_map[desc.opcode] = desc
      cls.opcode_encoding_map[desc.opcode] = desc.encoding

  @classmethod
  def decode_instruction(cls, logger, inst, core = None):
    logger.debug('%s.decode_instruction: inst=%s, core=%s', cls.__name__, inst, core)

    opcode = inst & 0x3F

    if opcode not in cls.opcode_desc_map:
      from ..cpu import InvalidOpcodeError
      raise InvalidOpcodeError(opcode, ip = core.current_ip if core is not None else None, core = core)

    return u32_to_encoding(inst, cls.opcode_encoding_map[opcode]), cls.opcode_desc_map[opcode], opcode

  @classmethod
  def disassemble_instruction(cls, logger, inst):
    logger.debug('%s.disassemble_instruction: inst=%s (%s)', cls.__name__, inst, inst.__class__.__name__)

    if isinstance(inst, ctypes.LittleEndianStructure):
      inst, desc = inst, cls.opcode_desc_map[inst.opcode]

    else:
      inst, desc, _ = cls.decode_instruction(logger, inst)

    mnemonic = desc.disassemble_mnemonic(inst)
    operands = desc.disassemble_operands(logger, inst)

    return (mnemonic + ' ' + ', '.join(operands)) if operands else mnemonic

#
# Main instruction set
#

def RI_VAL(core, inst, reg, sign_extend = True):
  if inst.immediate_flag == 1:
    if sign_extend is True:
      return inst.sign_extend_immediate(core.LOGGER, inst)

    return u32_t(0x00000000 | inst.immediate).value

  return core.registers.map[getattr(inst, reg)].value

def RI_ADDR(core, inst, reg):
  core.DEBUG('RI_ADDR: inst=%s, reg=%s', inst, reg)

  base = core.registers.map[reg].value
  offset = inst.sign_extend_immediate(core.LOGGER, inst) if inst.immediate_flag == 1 else 0

  return u32_t(base + offset).value

def JUMP(core, inst):
  core.DEBUG('JUMP: inst=%s', inst)
  core.DEBUG('  IP=%s', UINT32_FMT(core.registers.ip.value))

  if inst.immediate_flag == 0:
    core.DEBUG('  register=%d, value=%s', inst.reg, UINT32_FMT(core.registers.map[inst.reg].value))
    core.registers.ip.value = core.registers.map[inst.reg].value

  else:
    v = inst.sign_extend_immediate(core.LOGGER, inst)
    nip = u32_t(core.registers.ip.value)
    nip.value += (v << 2)
    core.DEBUG('  offset=%s, aligned=%s, ip=%s, new=%s', UINT32_FMT(v), UINT32_FMT(v << 2), UINT32_FMT(core.registers.ip.value), UINT32_FMT(nip.value))
    core.registers.ip.value = nip.value

  core.DEBUG('JUMP: new ip=%s', UINT32_FMT(core.registers.ip.value))

def update_arith_flags(core, reg):
  """
  Set relevant arithmetic flags according to content of registers. Flags are set to zero at the beginning,
  then content of each register is examined, and ``S`` and ``Z`` flags are set.

  ``E`` flag is not touched, ``O`` flag is set to zero.

  :param u32_t reg: register
  """

  core.arith_zero = False
  core.arith_overflow = False
  core.arith_sign = False

  if reg.value == 0:
    core.arith_zero = True

  if reg.value & 0x80000000 != 0:
    core.arith_sign = True

class DuckyOpcodes(enum.IntEnum):
  NOP    =  0

  # Memory load/store
  LW     =  1
  LS     =  2
  LB     =  3
  STW    =  4
  STS    =  5
  STB    =  6
  CAS    =  7
  LA     =  8

  LI     =  9
  LIU    = 10
  MOV    = 11
  SWP    = 12

  INT    = 13
  RETINT = 14

  CALL   = 15
  RET    = 16

  CLI    = 17
  STI    = 18
  RST    = 19
  HLT    = 20
  IDLE   = 21
  LPM    = 22
  IPI    = 23

  PUSH   = 24
  POP    = 25

  INC    = 26
  DEC    = 27
  ADD    = 28
  SUB    = 29
  MUL    = 30
  DIV    = 31
  UDIV   = 32
  MOD    = 33

  AND    = 34
  OR     = 35
  XOR    = 36
  NOT    = 37
  SHIFTL = 38
  SHIFTR = 39

  OUTW   = 40
  OUTS   = 41
  OUTB   = 42
  INW    = 43
  INS    = 44
  INB    = 45

  # Branch instructions
  J      = 46

  # Condition instructions
  CMP    = 47
  CMPU   = 48
  SET    = 49
  BRANCH = 50

  # Control instructions
  CTR    = 60
  CTW    = 61
  FPTC   = 62
  SIS    = 63


class NOP(Descriptor):
  mnemonic = 'nop'
  opcode = DuckyOpcodes.NOP
  encoding = EncodingI

  @staticmethod
  def execute(core, inst):
    pass


#
# Interrupts
#
class INT(Descriptor_RI):
  mnemonic      = 'int'
  opcode        = DuckyOpcodes.INT

  @staticmethod
  def execute(core, inst):
    core.do_int(RI_VAL(core, inst, 'reg'))

class IPI(Descriptor_R_RI):
  mnemonic = 'ipi'
  opcode = DuckyOpcodes.IPI

  @staticmethod
  def execute(core, inst):
    core.check_protected_ins()

    cpuid = core.registers.map[inst.reg1].value
    cpuid, coreid = cpuid >> 16, cpuid & 0xFFFF

    core.cpu.machine.cpus[cpuid].cores[coreid].irq(RI_VAL(core, inst, 'reg2'))

class RETINT(Descriptor):
  mnemonic = 'retint'
  opcode   = DuckyOpcodes.RETINT
  encoding = EncodingI

  @staticmethod
  def execute(core, inst):
    core.check_protected_ins()
    core.exit_interrupt()

#
# Jumps
#
class _JUMP(Descriptor):
  operands = 'ri'
  encoding = EncodingI
  relative_address = True
  inst_aligned = True

  @staticmethod
  def assemble_operands(logger, buffer, inst, operands):
    from .assemble import Reference

    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    if 'register_n0' in operands:
      ENCODE(logger, buffer, inst, 'reg', 5, operands['register_n0'])

    else:
      v = operands['immediate']

      ENCODE(logger, buffer, inst, 'immediate_flag', 1, 1)

      if isinstance(v, integer_types):
        if v & 0x3 != 0:
          raise buffer.get_error(UnalignedJumpTargetError, 'address=%s' % UINT32_FMT(v))

        ENCODE(logger, buffer, inst, 'immediate', 20, v >> 2)

      elif isinstance(v, string_types):
        inst.refers_to = Reference(label = v)

  @staticmethod
  def disassemble_operands(logger, inst):
    if inst.immediate_flag == 0:
      return [REGISTER_NAMES[inst.reg]]

    return [str(inst.refers_to) if hasattr(inst, 'refers_to') and inst.refers_to is not None else UINT32_FMT(inst.immediate << 2)]

class CALL(_JUMP):
  mnemonic      = 'call'
  opcode        = DuckyOpcodes.CALL

  @staticmethod
  def execute(core, inst):
    core.create_frame()

    JUMP(core, inst)

    if core.check_frames:
      core.frames[-1].IP = core.registers.ip.value

  @staticmethod
  def jit(core, inst):
    ip = core.registers.ip
    sp = core.registers.sp
    fp = core.registers.fp
    push = core.raw_push

    if inst.immediate_flag == 0:
      reg = core.registers.map[inst.reg]

      def __jit_call():
        push(ip.value)
        push(fp.value)
        fp.value = sp.value
        ip.value = reg.value

      return __jit_call

    else:
      i = inst.sign_extend_immediate(core.LOGGER, inst) << 2

      def __jit_call():
        push(ip.value)
        push(fp.value)
        fp.value = sp.value
        ip.value += i

      return __jit_call

class J(_JUMP):
  mnemonic = 'j'
  opcode   = DuckyOpcodes.J

  @staticmethod
  def execute(core, inst):
    JUMP(core, inst)

  @staticmethod
  def jit(core, inst):
    ip = core.registers.ip

    if inst.immediate_flag == 0:
      reg = core.registers.map[inst.reg]

      def __jit_j():
        ip.value = reg.value

      return __jit_j

    else:
      i = inst.sign_extend_immediate(core.LOGGER, inst) << 2

      def __jit_j():
        ip.value += i

      return __jit_j

class RET(Descriptor):
  mnemonic = 'ret'
  opcode   = DuckyOpcodes.RET
  encoding = EncodingI

  @staticmethod
  def execute(core, inst):
    core.destroy_frame()

  @staticmethod
  def jit(core, inst):
    ip = core.registers.ip
    fp = core.registers.fp
    pop = core.raw_pop

    def __jit_ret():
      fp.value = pop()
      ip.value = pop()

    return __jit_ret

#
# CPU
#
class LPM(Descriptor):
  mnemonic = 'lpm'
  opcode = DuckyOpcodes.LPM
  encoding = EncodingI

  @staticmethod
  def execute(core, inst):
    core.check_protected_ins()
    core.privileged = False

class CLI(Descriptor):
  mnemonic = 'cli'
  opcode = DuckyOpcodes.CLI
  encoding = EncodingI

  @staticmethod
  def execute(core, inst):
    core.check_protected_ins()
    core.hwint_allowed = False

class STI(Descriptor):
  mnemonic = 'sti'
  opcode = DuckyOpcodes.STI
  encoding = EncodingI

  @staticmethod
  def execute(core, inst):
    core.check_protected_ins()
    core.hwint_allowed = True

class HLT(Descriptor_RI):
  mnemonic = 'hlt'
  opcode = DuckyOpcodes.HLT

  @staticmethod
  def execute(core, inst):
    core.check_protected_ins()
    core.exit_code = RI_VAL(core, inst, 'reg')
    core.halt()

class RST(Descriptor):
  mnemonic = 'rst'
  opcode = DuckyOpcodes.RST
  encoding = EncodingI

  @staticmethod
  def execute(core, inst):
    core.check_protected_ins()
    core.reset()

class IDLE(Descriptor):
  mnemonic = 'idle'
  opcode = DuckyOpcodes.IDLE
  encoding = EncodingI

  @staticmethod
  def execute(core, inst):
    core.change_runnable_state(idle = True)

class SIS(Descriptor_RI):
  mnemonic = 'sis'
  opcode = DuckyOpcodes.SIS

  @staticmethod
  def execute(core, inst):
    core.instruction_set = get_instruction_set(RI_VAL(core, inst, 'reg'))


#
# Stack
#
class PUSH(Descriptor_RI):
  mnemonic = 'push'
  opcode = DuckyOpcodes.PUSH

  @staticmethod
  def execute(core, inst):
    core.raw_push(RI_VAL(core, inst, 'reg'))

  @staticmethod
  def jit(core, inst):
    push = core.raw_push

    if inst.immediate_flag == 0:
      reg = core.registers.map[inst.reg]

      def __jit_push():
        push(reg.value)

      return __jit_push

    else:
      i = inst.sign_extend_immediate(core.LOGGER, inst)

      def __jit_push():
        push(i)

      return __jit_push

class POP(Descriptor_R):
  mnemonic = 'pop'
  opcode = DuckyOpcodes.POP

  @staticmethod
  def execute(core, inst):
    core.pop(inst.reg1)
    update_arith_flags(core, core.registers.map[inst.reg1])

  @staticmethod
  def jit(core, inst):
    pop = core.raw_pop
    reg = core.registers.map[inst.reg1]

    def __jit_pop():
      reg.value = v = pop()
      core.arith_zero = v == 0
      core.arith_overflow = False
      core.arith_sign = (v & 0x80000000) != 0

    return __jit_pop

#
# Arithmetic
#
class INC(Descriptor_R):
  mnemonic = 'inc'
  opcode = DuckyOpcodes.INC

  @staticmethod
  def execute(core, inst):
    core.registers.map[inst.reg1].value += 1
    update_arith_flags(core, core.registers.map[inst.reg1])
    core.arith_overflow = core.registers.map[inst.reg1].value == 0

  @staticmethod
  def jit(core, inst):
    reg = core.registers.map[inst.reg1]

    def __jit_inc():
      old, new = reg.value, reg.value + 1
      reg.value = new

      if old == 0:
        core.arith_zero = core.arith_overflow = core.arith_sign = False

      elif old == 0xFFFFFFFF:
        core.arith_zero = core.arith_overflow = True
        core.arith_sign = False

      else:
        core.arith_zero = False
        core.arith_overflow = False
        core.arith_sign = (new & 0x80000000) != 0

    return __jit_inc

class DEC(Descriptor_R):
  mnemonic = 'dec'
  opcode = DuckyOpcodes.DEC

  @staticmethod
  def execute(core, inst):
    core.registers.map[inst.reg1].value -= 1
    update_arith_flags(core, core.registers.map[inst.reg1])

  @staticmethod
  def jit(core, inst):
    reg = core.registers.map[inst.reg1]

    def __jit_dec():
      old, new = reg.value, reg.value - 1
      reg.value = new

      if old == 0:
        core.arith_zero = core.arith_overflow = False
        core.arith_sign = True

      elif old == 1:
        core.arith_zero = True
        core.arith_overflow = core.arith_sign = False

      else:
        core.arith_zero = core.arith_overflow = False
        core.arith_sign = (new & 0x80000000) != 0

    return __jit_dec

class _BINOP(Descriptor_R_RI):
  encoding = EncodingR

  @staticmethod
  def execute(core, inst):
    r = core.registers.map[inst.reg1]
    v = RI_VAL(core, inst, 'reg2')

    if inst.opcode == DuckyOpcodes.ADD:
      v = r.value + v

    elif inst.opcode == DuckyOpcodes.SUB:
      v = r.value - v

    elif inst.opcode == DuckyOpcodes.MUL:
      x = i32_t(r.value).value
      y = i32_t(v).value
      v = x * y

    elif inst.opcode == DuckyOpcodes.DIV:
      x = i32_t(r.value).value
      y = i32_t(v).value

      if abs(y) > abs(x):
        v = 0

      else:
        v = x // y

    elif inst.opcode == DuckyOpcodes.UDIV:
      x = u32_t(r.value).value
      y = u32_t(v).value

      v = x // y

    elif inst.opcode == DuckyOpcodes.MOD:
      x = i32_t(r.value).value
      y = i32_t(v).value

      v = x % y

    r.value = v
    update_arith_flags(core, r)

    if v > 0xFFFFFFFF:
      core.arith_overflow = True

class ADD(_BINOP):
  mnemonic = 'add'
  opcode = DuckyOpcodes.ADD

  @staticmethod
  def jit(core, inst):
    core.DEBUG('JIT: %s', inst)

    if inst.immediate_flag == 1:
      reg = core.registers.map[inst.reg1]
      i = inst.sign_extend_immediate(core.LOGGER, inst)

      def __jit_add():
        reg.value = v = reg.value + i
        core.arith_zero = reg.value == 0
        core.arith_overflow = v > 0xFFFFFFFF
        core.arith_sign = (v & 0x80000000) != 0

      return __jit_add

    else:
      reg1 = core.registers.map[inst.reg1]
      reg2 = core.registers.map[inst.reg2]

      def __jit_add():
        reg1.value = v = reg1.value + reg2.value
        core.arith_zero = reg1.value == 0
        core.arith_overflow = v > 0xFFFFFFFF
        core.arith_sign = (v & 0x80000000) != 0

      return __jit_add

class SUB(_BINOP):
  mnemonic = 'sub'
  opcode = DuckyOpcodes.SUB

  @staticmethod
  def jit(core, inst):
    core.DEBUG('JIT: %s', inst)

    if inst.immediate_flag == 1:
      reg = core.registers.map[inst.reg1]
      i = inst.sign_extend_immediate(core.LOGGER, inst)

      def __jit_sub():
        reg.value = v = reg.value - i
        core.arith_zero = reg.value == 0
        core.arith_overflow = v > 0xFFFFFFFF
        core.arith_sign = (v & 0x80000000) != 0

      return __jit_sub

    else:
      reg1 = core.registers.map[inst.reg1]
      reg2 = core.registers.map[inst.reg2]

      def __jit_sub():
        reg1.value = v = reg1.value - reg2.value
        core.arith_zero = reg1.value == 0
        core.arith_overflow = v > 0xFFFFFFFF
        core.arith_sign = (v & 0x80000000) != 0

      return __jit_sub

class MUL(_BINOP):
  mnemonic = 'mul'
  opcode = DuckyOpcodes.MUL

  @staticmethod
  def jit(core, inst):
    core.DEBUG('JIT: %s', inst)

    if inst.immediate_flag == 1:
      reg = core.registers.map[inst.reg1]
      i = i32_t(inst.sign_extend_immediate(core.LOGGER, inst)).value

      def __jit_mul():
        reg.value = v = i32_t(reg.value).value * i
        core.arith_zero = reg.value == 0
        core.arith_overflow = v > 0xFFFFFFFF
        core.arith_sign = (v & 0x80000000) != 0

      return __jit_mul

    else:
      reg1 = core.registers.map[inst.reg1]
      reg2 = core.registers.map[inst.reg2]

      def __jit_mul():
        reg1.value = v = i32_t(reg1.value).value * i32_t(reg2.value).value
        core.arith_zero = reg1.value == 0
        core.arith_overflow = v > 0xFFFFFFFF
        core.arith_sign = (v & 0x80000000) != 0

      return __jit_mul

class DIV(_BINOP):
  mnemonic = 'div'
  opcode = DuckyOpcodes.DIV

class UDIV(_BINOP):
  mnemonic = 'udiv'
  opcode = DuckyOpcodes.UDIV

class MOD(_BINOP):
  mnemonic = 'mod'
  opcode = DuckyOpcodes.MOD


#
# Conditional and unconditional jumps
#
class _COND(Descriptor):
  encoding = EncodingC

  FLAGS = ['arith_equal', 'arith_zero', 'arith_overflow', 'arith_sign', 'l', 'g']
  GFLAGS = [0, 1, 2, 3]
  MNEMONICS = ['e', 'z', 'o', 's', 'g', 'l']

  @staticmethod
  def set_condition(logger, buffer, inst, flag, value):
    logger.debug('set_condition: flag=%s, value=%s', flag, value)

    ENCODE(logger, buffer, inst, 'flag', 3, _COND.FLAGS.index(flag))
    ENCODE(logger, buffer, inst, 'value', 1, 1 if value is True else 0)

  @staticmethod
  def evaluate(core, inst):
    # genuine flags
    if inst.flag in _COND.GFLAGS and inst.value == getattr(core, _COND.FLAGS[inst.flag]):
      return True

    # "less than" flag
    if inst.flag == 4:
      if inst.value == 1 and core.arith_sign is True and core.arith_equal is not True:
        return True

      if inst.value == 0 and (core.arith_sign is not True or core.arith_equal is True):
        return True

    # "greater than" flag
    if inst.flag == 5:
      if inst.value == 1 and core.arith_sign is not True and core.arith_equal is not True:
        return True

      if inst.value == 0 and (core.arith_sign is True or core.arith_equal is True):
        return True

    return False

class _BRANCH(_COND):
  operands = 'ri'
  opcode = DuckyOpcodes.BRANCH
  relative_address = True
  inst_aligned = True

  @classmethod
  def assemble_operands(cls, logger, buffer, inst, operands):
    from .assemble import Reference

    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    if 'register_n0' in operands:
      ENCODE(logger, buffer, inst, 'reg', 5, operands['register_n0'])

    else:
      v = operands['immediate']

      ENCODE(logger, buffer, inst, 'immediate_flag', 1, 1)

      if isinstance(v, integer_types):
        if v & 0x3 != 0:
          raise buffer.get_error(UnalignedJumpTargetError, 'address=%s' % UINT32_FMT(v))

        ENCODE(logger, buffer, inst, 'immediate', 16, v >> 2)

      elif isinstance(v, string_types):
        inst.refers_to = Reference(label = v)

    set_condition = partial(_COND.set_condition, logger, buffer, inst)

    if cls is BE:
      set_condition('arith_equal', True)

    elif cls is BNE:
      set_condition('arith_equal', False)

    elif cls is BZ:
      set_condition('arith_zero', True)

    elif cls is BNZ:
      set_condition('arith_zero', False)

    elif cls is BO:
      set_condition('arith_overflow', True)

    elif cls is BNO:
      set_condition('arith_overflow', False)

    elif cls is BS:
      set_condition('arith_sign', True)

    elif cls is BNS:
      set_condition('arith_sign', False)

    elif cls is BL:
      set_condition('l', True)

    elif cls is BLE:
      set_condition('g', False)

    elif cls is BG:
      set_condition('g', True)

    elif cls is BGE:
      set_condition('l', False)

  @staticmethod
  def fill_reloc_slot(logger, inst, slot):
    inst.fill_reloc_slot(logger, inst, slot)

    slot.flags.inst_aligned = True

  @staticmethod
  def disassemble_operands(logger, inst):
    if inst.immediate_flag == 0:
      return [REGISTER_NAMES[inst.reg]]

    return [str(inst.refers_to) if hasattr(inst, 'refers_to') and inst.refers_to is not None else UINT16_FMT(inst.immediate << 2)]

  @staticmethod
  def disassemble_mnemonic(inst):
    if inst.flag in _COND.GFLAGS:
      return 'b%s%s' % ('n' if inst.value == 0 else '', _COND.MNEMONICS[inst.flag])

    else:
      if inst.flag == _COND.FLAGS.index('l'):
        return 'bl' if inst.value == 1 else 'bge'

      elif inst.flag == _COND.FLAGS.index('g'):
        return 'bg' if inst.value == 1 else 'ble'

  @staticmethod
  def execute(core, inst):
    if _COND.evaluate(core, inst):
      JUMP(core, inst)

  @staticmethod
  def jit(core, inst):
    core.DEBUG('JIT: %s', inst)

    if inst.immediate_flag == 1:
      i = inst.sign_extend_immediate(core.LOGGER, inst) << 2

    else:
      reg = core.registers.map[inst.reg]

    ip = core.registers.ip

    if inst.flag in _COND.GFLAGS:
      if inst.flag == 0:
        if inst.value == 0:
          if inst.immediate_flag == 0:
            def __branch_ne():
              if core.arith_equal is False:
                ip.value = reg.value

            return __branch_ne

          else:
            def __branch_ne():
              if core.arith_equal is False:
                ip.value += i

            return __branch_ne

        else:
          if inst.immediate_flag == 0:
            def __branch_e():
              if core.arith_equal is True:
                ip.value = reg.value

            return __branch_e

          else:
            def __branch_e():
              if core.arith_equal is True:
                ip.value += i

            return __branch_e

      elif inst.flag == 1:
        if inst.value == 0:
          if inst.immediate_flag == 0:
            def __branch_nz():
              if core.arith_zero is False:
                ip.value = reg.value

            return __branch_nz

          else:
            def __branch_nz():
              if core.arith_zero is False:
                ip.value += i

            return __branch_nz

        else:
          if inst.immediate_flag == 0:
            def __branch_z():
              if core.arith_zero is True:
                ip.value = reg.value

            return __branch_z

          else:
            def __branch_z():
              if core.arith_zero is True:
                ip.value += i

            return __branch_z

      elif inst.flag == 2:
        if inst.value == 0:
          if inst.immediate_flag == 0:
            def __branch_no():
              if core.arith_overflow is False:
                ip.value = reg.value

            return __branch_no

          else:
            def __branch_no():
              if core.arith_overflow is False:
                ip.value += i

            return __branch_no

        else:
          if inst.immediate_flag == 0:
            def __branch_o():
              if core.arith_overflow is True:
                ip.value = reg.value

            return __branch_o

          else:
            def __branch_o():
              if core.arith_overflow is True:
                ip.value += i

            return __branch_o

      elif inst.flag == 3:
        if inst.value == 0:
          if inst.immediate_flag == 0:
            def __branch_ns():
              if core.arith_sign is False:
                ip.value = reg.value

            return __branch_ns

          else:
            def __branch_ns():
              if core.arith_sign is False:
                ip.value += i

            return __branch_ns

        else:
          if inst.immediate_flag == 0:
            def __branch_s():
              if core.arith_sign is True:
                ip.value = reg.value

            return __branch_s

          else:
            def __branch_s():
              if core.arith_sign is True:
                ip.value += i

            return __branch_s

    return None

class _SET(_COND):
  operands = 'r'
  opcode = DuckyOpcodes.SET

  @classmethod
  def assemble_operands(cls, logger, buffer, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    ENCODE(logger, buffer, inst, 'reg', 5, operands['register_n0'])

    set_condition = partial(_COND.set_condition, logger, buffer, inst)

    if cls is SETE:
      set_condition('arith_equal', True)

    elif cls is SETNE:
      set_condition('arith_equal', False)

    elif cls is SETZ:
      set_condition('arith_zero', True)

    elif cls is SETNZ:
      set_condition('arith_zero', False)

    elif cls is SETO:
      set_condition('arith_overflow', True)

    elif cls is SETNO:
      set_condition('arith_overflow', False)

    elif cls is SETS:
      set_condition('arith_sign', True)

    elif cls is SETNS:
      set_condition('arith_sign', False)

    elif cls is SETL:
      set_condition('l', True)

    elif cls is SETLE:
      set_condition('g', False)

    elif cls is SETG:
      set_condition('g', True)

    elif cls is SETGE:
      set_condition('l', False)

  @staticmethod
  def disassemble_operands(logger, inst):
    return [REGISTER_NAMES[inst.reg]]

  @staticmethod
  def disassemble_mnemonic(inst):
    if inst.flag in _COND.GFLAGS:
      return 'set%s%s' % ('n' if inst.value == 0 else '', _COND.MNEMONICS[inst.flag])

    else:
      return 'set%s%s' % (_COND.MNEMONICS[inst.flag], 'e' if inst.value == 1 else '')

  @staticmethod
  def execute(core, inst):
    core.registers.map[inst.reg].value = 1 if _COND.evaluate(core, inst) is True else 0
    update_arith_flags(core, core.registers.map[inst.reg])

class _CMP(Descriptor_R_RI):
  encoding = EncodingR

  @staticmethod
  def evaluate(core, x, y, signed = True):
    """
    Compare two numbers, and update relevant flags. Signed comparison is used unless ``signed`` is ``False``.
    All arithmetic flags are set to zero before the relevant ones are set.

    ``O`` flag is reset like the others, therefore caller has to take care of it's setting if it's required
    to set it.

    :param u32 x: left hand number
    :param u32 y: right hand number
    :param bool signed: use signed, defaults to ``True``
    """

    core.arith_equal = False
    core.arith_zero = False
    core.arith_overflow = False
    core.arith_sign = False

    if x == y:
      core.arith_equal = True

      if x == 0:
        core.arith_zero = True

      return

    if signed:
      x = i32_t(x).value
      y = i32_t(y).value

    if x < y:
      core.arith_sign = True

class CMP(_CMP):
  mnemonic = 'cmp'
  opcode = DuckyOpcodes.CMP

  @staticmethod
  def execute(core, inst):
    _CMP.evaluate(core, core.registers.map[inst.reg1].value, RI_VAL(core, inst, 'reg2'))

  @staticmethod
  def jit(core, inst):

    if inst.immediate_flag == 0:
      reg1 = core.registers.map[inst.reg1]
      reg2 = core.registers.map[inst.reg2]

      def __jit_cmp():
        x = reg1.value
        y = reg2.value

        core.arith_overflow = False

        if x == y:
          core.arith_equal = True
          core.arith_zero = x == 0
          core.arith_sign = False
          return

        core.arith_equal = False
        core.arith_zero  = False
        core.arith_sign = i32_t(x).value < i32_t(y).value

      return __jit_cmp

    else:
      reg = core.registers.map[inst.reg1]
      y = inst.sign_extend_immediate(core.LOGGER, inst)
      y_signed = i32_t(y).value

      def __jit_cmp():
        x = reg.value

        core.arith_overflow = False

        if x == y:
          core.arith_equal = True
          core.arith_zero = x == 0
          core.arith_sign = False
          return

        core.arith_equal = False
        core.arith_zero  = False
        core.arith_sign = i32_t(x).value < y_signed

      return __jit_cmp

    return None

class CMPU(_CMP):
  mnemonic = 'cmpu'
  opcode = DuckyOpcodes.CMPU

  @staticmethod
  def execute(core, inst):
    _CMP.evaluate(core, core.registers.map[inst.reg1].value, RI_VAL(core, inst, 'reg2', sign_extend = False), signed = False)

class BE(_BRANCH):
  mnemonic = 'be'

class BNE(_BRANCH):
  mnemonic = 'bne'

class BNS(_BRANCH):
  mnemonic = 'bns'

class BNZ(_BRANCH):
  mnemonic = 'bnz'

class BS(_BRANCH):
  mnemonic = 'bs'

class BZ(_BRANCH):
  mnemonic = 'bz'

class BO(_BRANCH):
  mnemonic = 'bo'

class BNO(_BRANCH):
  mnemonic = 'bno'

class BG(_BRANCH):
  mnemonic = 'bg'

class BGE(_BRANCH):
  mnemonic = 'bge'

class BL(_BRANCH):
  mnemonic = 'bl'

class BLE(_BRANCH):
  mnemonic = 'ble'

class SETE(_SET):
  mnemonic = 'sete'

class SETNE(_SET):
  mnemonic = 'setne'

class SETZ(_SET):
  mnemonic = 'setz'

class SETNZ(_SET):
  mnemonic = 'setnz'

class SETO(_SET):
  mnemonic = 'seto'

class SETNO(_SET):
  mnemonic = 'setno'

class SETS(_SET):
  mnemonic = 'sets'

class SETNS(_SET):
  mnemonic = 'setns'

class SETG(_SET):
  mnemonic = 'setg'

class SETGE(_SET):
  mnemonic = 'setge'

class SETL(_SET):
  mnemonic = 'setl'

class SETLE(_SET):
  mnemonic = 'setle'


#
# IO
#

class _IN(Descriptor_R_RI):
  encoding = EncodingR

  @staticmethod
  def execute(core, inst):
    port = RI_VAL(core, inst, 'reg2')

    core.check_protected_port(port)

    r = core.registers.map[inst.reg1]
    dev = core.cpu.machine.ports[port]

    if inst.opcode == DuckyOpcodes.INB:
      r.value = dev.read_u8(port)

    elif inst.opcode == DuckyOpcodes.INS:
      r.value = dev.read_u16(port)

    else:
      r.value = dev.read_u32(port)

    update_arith_flags(core, r)

  @staticmethod
  def jit(core, inst):
    reg = core.registers.map[inst.reg1]

    if inst.immediate_flag == 1:
      port = inst.sign_extend_immediate(core.LOGGER, inst)
      dev = core.cpu.machine.ports[port]
      is_protected = partial(dev.is_port_protected, port)

      if inst.opcode == DuckyOpcodes.INB:
        reader = dev.read_u8

        def __jit_inb():
          if not core.privileged and is_protected():
            raise AccessViolationError('Access to port not allowed in unprivileged mode: inst={}, port={}'.format(inst, port))

          reg.value = v = reader(port)
          core.arith_zero = v == 0
          core.arith_overflow = core.arith_sign = False

        return __jit_inb

    return None

class INB(_IN):
  mnemonic = 'inb'
  opcode   = DuckyOpcodes.INB

class INS(_IN):
  mnemonic = 'ins'
  opcode   = DuckyOpcodes.INS

class INW(_IN):
  mnemonic = 'inw'
  opcode   = DuckyOpcodes.INW

class _OUT(Descriptor_RI_R):
  encoding = EncodingR

  @staticmethod
  def execute(core, inst):
    port = RI_VAL(core, inst, 'reg1')

    core.check_protected_port(port)

    r = core.registers.map[inst.reg2]
    dev = core.cpu.machine.ports[port]

    if inst.opcode == DuckyOpcodes.OUTB:
      dev.write_u8(port, r.value & 0xFF)

    elif inst.opcode == DuckyOpcodes.OUTS:
      dev.write_u16(port, r.value & 0xFFFF)

    else:
      dev.write_u32(port, r.value)

  @staticmethod
  def jit(core, inst):
    reg = core.registers.map[inst.reg2]

    if inst.immediate_flag == 1:
      port = inst.sign_extend_immediate(core.LOGGER, inst)
      dev = core.cpu.machine.ports[port]
      is_protected = partial(dev.is_port_protected, port)

      if inst.opcode == DuckyOpcodes.OUTB:
        writer = dev.write_u8

        def __jit_outb():
          if not core.privileged and is_protected():
            raise AccessViolationError('Access to port not allowed in unprivileged mode: inst={}, port={}'.format(inst, port))

          writer(port, reg.value & 0xFF)

        return __jit_outb

    return None

class OUTB(_OUT):
  mnemonic      = 'outb'
  opcode = DuckyOpcodes.OUTB

class OUTW(_OUT):
  mnemonic = 'outw'
  opcode   = DuckyOpcodes.OUTW

class OUTS(_OUT):
  mnemonic = 'outs'
  opcode   = DuckyOpcodes.OUTS


#
# Bit operations
#
class _BITOP(Descriptor_R_RI):
  encoding = EncodingR

  @staticmethod
  def execute(core, inst):
    r = core.registers.map[inst.reg1]
    v = RI_VAL(core, inst, 'reg2')

    if inst.opcode == DuckyOpcodes.AND:
      value = r.value & v

    elif inst.opcode == DuckyOpcodes.OR:
      value = r.value | v

    elif inst.opcode == DuckyOpcodes.XOR:
      value = r.value ^ v

    elif inst.opcode == DuckyOpcodes.SHIFTL:
      value = r.value << min(v, 32)

    elif inst.opcode == DuckyOpcodes.SHIFTR:
      value = r.value >> min(v, 32)

    r.value = value
    update_arith_flags(core, r)

    if value > 0xFFFFFFFF:
      core.arith_overflow = True

class AND(_BITOP):
  mnemonic = 'and'
  opcode = DuckyOpcodes.AND

  @staticmethod
  def jit(core, inst):
    if inst.immediate_flag == 1:
      reg = core.registers.map[inst.reg1]
      i = inst.sign_extend_immediate(core.LOGGER, inst)

      def __jit_and():
        reg.value = v = reg.value & i
        core.arith_zero = v == 0
        core.arith_overflow = False
        core.arith_sign = (v & 0x80000000) != 0

      return __jit_and

    else:
      reg1 = core.registers.map[inst.reg1]
      reg2 = core.registers.map[inst.reg2]

      def __jit_and():
        reg1.value = v = reg1.value & reg2.value
        core.arith_zero = v == 0
        core.arith_overflow = False
        core.arith_sign = (v & 0x80000000) != 0

      return __jit_and

class OR(_BITOP):
  mnemonic = 'or'
  opcode = DuckyOpcodes.OR

  @staticmethod
  def jit(core, inst):
    if inst.immediate_flag == 1:
      reg = core.registers.map[inst.reg1]
      i = inst.sign_extend_immediate(core.LOGGER, inst)

      def __jit_or():
        reg.value = v = reg.value | i
        core.arith_zero = v == 0
        core.arith_overflow = False
        core.arith_sign = (v & 0x80000000) != 0

      return __jit_or

    else:
      reg1 = core.registers.map[inst.reg1]
      reg2 = core.registers.map[inst.reg2]

      def __jit_or():
        reg1.value = v = reg1.value | reg2.value
        core.arith_zero = v == 0
        core.arith_overflow = False
        core.arith_sign = (v & 0x80000000) != 0

      return __jit_or

class XOR(_BITOP):
  mnemonic = 'xor'
  opcode = DuckyOpcodes.XOR

  @staticmethod
  def jit(core, inst):
    if inst.immediate_flag == 1:
      reg = core.registers.map[inst.reg1]
      i = inst.sign_extend_immediate(core.LOGGER, inst)

      def __jit_xor():
        reg.value = v = reg.value ^ i
        core.arith_zero = v == 0
        core.arith_overflow = False
        core.arith_sign = (v & 0x80000000) != 0

      return __jit_xor

    else:
      reg1 = core.registers.map[inst.reg1]
      reg2 = core.registers.map[inst.reg2]

      def __jit_xor():
        reg1.value = v = reg1.value ^ reg2.value
        core.arith_zero = v == 0
        core.arith_overflow = False
        core.arith_sign = (v & 0x80000000) != 0

      return __jit_xor

class SHIFTL(_BITOP):
  mnemonic = 'shiftl'
  opcode = DuckyOpcodes.SHIFTL

  @staticmethod
  def jit(core, inst):
    if inst.immediate_flag == 1:
      reg = core.registers.map[inst.reg1]
      i = min(inst.sign_extend_immediate(core.LOGGER, inst), 32)

      if i == 0:
        def __jit_shiftl():
          v = reg.value
          core.arith_zero = v == 0
          core.arith_overflow = False
          core.arith_sign = (v & 0x80000000) != 0

        return __jit_shiftl

      elif i == 32:
        def __jit_shiftl():
          v = reg.value << i
          reg.value = 0
          core.arith_zero = True
          core.arith_overflow = (v & ~0xFFFFFFFF) != 0
          core.arith_sign = False

        return __jit_shiftl

      else:
        def __jit_shiftl():
          v = reg.value << i
          reg.value = v
          core.arith_zero = (v & 0xFFFFFFFF) == 0
          core.arith_overflow = (v & ~0xFFFFFFFF) != 0
          core.arith_sign = (v & 0x80000000) != 0

        return __jit_shiftl

    else:
      reg1 = core.registers.map[inst.reg1]
      reg2 = core.registers.map[inst.reg2]

      def __jit_shiftl():
        i = min(reg2.value, 32)

        if i == 0:
          v = reg1.value
          core.arith_zero = v == 0
          core.arith_overflow = False
          core.arith_sign = (v & 0x80000000) != 0

        elif i == 32:
          v = reg1.value << i
          reg1.value = 0
          core.arith_zero = True
          core.arith_overflow = (v & ~0xFFFFFFFF) != 0
          core.arith_sign = False

        else:
          v = reg1.value << i
          reg1.value = v
          core.arith_zero = (v & 0xFFFFFFFF) == 0
          core.arith_overflow = (v & ~0xFFFFFFFF) != 0
          core.arith_sign = (v & 0x80000000) != 0

      return __jit_shiftl

class SHIFTR(_BITOP):
  mnemonic = 'shiftr'
  opcode = DuckyOpcodes.SHIFTR

  @staticmethod
  def jit(core, inst):
    if inst.immediate_flag == 1:
      reg = core.registers.map[inst.reg1]
      i = min(inst.sign_extend_immediate(core.LOGGER, inst), 32)

      if i == 0:
        def __jit_shiftr():
          v = reg.value
          core.arith_zero = v == 0
          core.arith_overflow = False
          core.arith_sign = (v & 0x80000000) != 0

        return __jit_shiftr

      elif i == 32:
        def __jit_shiftr():
          reg.value = 0
          core.arith_zero = True
          core.arith_overflow = core.arith_sign = False

        return __jit_shiftr

      else:
        def __jit_shiftr():
          reg.value = v = reg.value >> i
          core.arith_zero = v == 0
          core.arith_overflow = False
          core.arith_sign = False

        return __jit_shiftr

    else:
      reg1 = core.registers.map[inst.reg1]
      reg2 = core.registers.map[inst.reg2]

      def __jit_shiftr():
        i = min(reg2.value, 32)

        if i == 0:
          v = reg1.value
          core.arith_zero = v == 0
          core.arith_overflow = False
          core.arith_sign = (v & 0x80000000) != 0

        elif i == 32:
          reg1.value = 0
          core.arith_zero = True
          core.arith_overflow = core.arith_sign = False

        else:
          reg1.value = v = reg1.value >> i
          core.arith_zero = v == 0
          core.arith_overflow = False
          core.arith_sign = False

      return __jit_shiftr

class NOT(Descriptor_R):
  mnemonic = 'not'
  opcode = DuckyOpcodes.NOT
  encoding = EncodingR

  @staticmethod
  def execute(core, inst):
    r = core.registers.map[inst.reg1]

    r.value = ~r.value
    update_arith_flags(core, r)


#
# Memory load/store operations
#
class CAS(Descriptor):
  mnemonic = 'cas'
  operands = 'r,r,r'
  opcode = DuckyOpcodes.CAS
  encoding = EncodingA

  @staticmethod
  def assemble_operands(logger, buffer, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    ENCODE(logger, buffer, inst, 'reg1', 5, operands['register_n0'])
    ENCODE(logger, buffer, inst, 'reg2', 5, operands['register_n1'])
    ENCODE(logger, buffer, inst, 'reg3', 5, operands['register_n2'])

  @staticmethod
  def disassemble_operands(logger, inst):
    return [
      REGISTER_NAMES[inst.reg1],
      REGISTER_NAMES[inst.reg2],
      REGISTER_NAMES[inst.reg3]
    ]

  @staticmethod
  def execute(core, inst):
    core.arith_equal = False

    reg1, reg2, reg3 = core.registers.map[inst.reg1], core.registers.map[inst.reg2], core.registers.map[inst.reg3]
    core.DEBUG('CAS.execute: reg1=%s (%s), reg2=%s (%s), reg3=%s (%s)', inst.reg1, UINT32_FMT(reg1.value), inst.reg2, UINT32_FMT(reg2.value), inst.reg3, UINT32_FMT(reg3.value))

    addr = core.registers.map[inst.reg1].value
    actual_value = core.MEM_IN32(reg1.value)

    core.DEBUG('CAS.execute: value=%s', UINT32_FMT(actual_value))

    if actual_value == reg2.value:
      core.MEM_OUT32(addr, reg3.value)
      core.arith_equal = True

    else:
      reg2.value = actual_value

class _LOAD(Descriptor):
  operands = 'r,a'
  encoding = EncodingR

  @staticmethod
  def assemble_operands(logger, buffer, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    ENCODE(logger, buffer, inst, 'reg1', 5, operands['register_n0'])
    ENCODE(logger, buffer, inst, 'reg2', 5, operands['areg'])

    offset_immediate = operands.get('offset_immediate')
    if offset_immediate is not None:
      ENCODE(logger, buffer, inst, 'immediate_flag', 1, 1)
      ENCODE(logger, buffer, inst, 'immediate', 15, int(offset_immediate))

  @staticmethod
  def disassemble_operands(logger, inst):
    operands = [REGISTER_NAMES[inst.reg1]]

    if inst.immediate_flag == 1:
      operands.append('%s[%s]' % (REGISTER_NAMES[inst.reg2], inst.sign_extend_immediate(logger, inst)))

    else:
      operands.append(REGISTER_NAMES[inst.reg2])

    return operands

  @staticmethod
  def execute(core, inst):
    r = core.registers.map[inst.reg1]
    addr = RI_ADDR(core, inst, inst.reg2)

    if inst.opcode == DuckyOpcodes.LW:
      r.value = core.MEM_IN32(addr)

    elif inst.opcode == DuckyOpcodes.LS:
      r.value = core.MEM_IN16(addr)

    else:
      r.value = core.MEM_IN8(addr)

    update_arith_flags(core, r)

  @staticmethod
  def jit(core, inst):
    reg1 = core.registers.map[inst.reg1]
    reg2 = core.registers.map[inst.reg2]

    if inst.opcode == DuckyOpcodes.LW:
      reader = core.MEM_IN32

      if inst.immediate_flag == 1:
        offset = inst.sign_extend_immediate(core.LOGGER, inst)

        if offset == 0:
          def __jit_lw():
            reg1.value = v = reader(reg2.value)
            core.arith_zero = v == 0
            core.arith_overflow = False
            core.arith_sign = (v & 0x80000000) != 0

          return __jit_lw

        else:
          def __jit_lw():
            reg1.value = v = reader((reg2.value + offset) % 4294967296)
            core.arith_zero = v == 0
            core.arith_overflow = False
            core.arith_sign = (v & 0x80000000) != 0

          return __jit_lw

      else:
        def __jit_lw():
          reg1.value = v = reader(reg2.value)
          core.arith_zero = v == 0
          core.arith_overflow = False
          core.arith_sign = (v & 0x80000000) != 0

        return __jit_lw

    elif inst.opcode == DuckyOpcodes.LS:
      reader = core.MEM_IN16

      if inst.immediate_flag == 1:
        offset = inst.sign_extend_immediate(core.LOGGER, inst)

        if offset == 0:
          def __jit_ls():
            reg1.value = v = reader(reg2.value)
            core.arith_zero = v == 0
            core.arith_overflow = False
            core.arith_sign = False

          return __jit_ls

        else:
          def __jit_ls():
            reg1.value = v = reader((reg2.value + offset) % 4294967296)
            core.arith_zero = v == 0
            core.arith_overflow = False
            core.arith_sign = False

          return __jit_ls

      else:
        def __jit_ls():
          reg1.value = v = reader(reg2.value)
          core.arith_zero = v == 0
          core.arith_overflow = False
          core.arith_sign = False

        return __jit_ls

    elif inst.opcode == DuckyOpcodes.LB:
      reader = core.MEM_IN8

      if inst.immediate_flag == 1:
        offset = inst.sign_extend_immediate(core.LOGGER, inst)

        if offset == 0:
          def __jit_lb():
            reg1.value = v = reader(reg2.value)
            core.arith_zero = v == 0
            core.arith_overflow = False
            core.arith_sign = False

          return __jit_lb

        else:
          def __jit_lb():
            reg1.value = v = reader((reg2.value + offset) % 4294967296)
            core.arith_zero = v == 0
            core.arith_overflow = False
            core.arith_sign = False

          return __jit_lb

      else:
        def __jit_lb():
          reg1.value = v = reader(reg2.value)
          core.arith_zero = v == 0
          core.arith_overflow = False
          core.arith_sign = False

        return __jit_lb

    return None

class _STORE(Descriptor):
  operands = 'a,r'
  encoding = EncodingR

  @staticmethod
  def assemble_operands(logger, buffer, inst, operands):
    logger.debug('assemble_operands: inst=%s, operands=%s', inst, operands)

    ENCODE(logger, buffer, inst, 'reg1', 5, operands['areg'])
    ENCODE(logger, buffer, inst, 'reg2', 5, operands['register_n1'])

    offset_immediate = operands.get('offset_immediate')
    if offset_immediate is not None:
      ENCODE(logger, buffer, inst, 'immediate_flag', 1, 1)
      ENCODE(logger, buffer, inst, 'immediate', 15, int(offset_immediate))

  @staticmethod
  def disassemble_operands(logger, inst):
    operands = []

    if inst.immediate_flag == 1:
      operands.append('%s[%s]' % (REGISTER_NAMES[inst.reg1], inst.sign_extend_immediate(logger, inst)))

    else:
      operands.append(REGISTER_NAMES[inst.reg1])

    operands.append(REGISTER_NAMES[inst.reg2])

    return operands

  @staticmethod
  def execute(core, inst):
    addr = RI_ADDR(core, inst, inst.reg1)

    if inst.opcode == DuckyOpcodes.STW:
      core.MEM_OUT32(addr, core.registers.map[inst.reg2].value)

    elif inst.opcode == DuckyOpcodes.STS:
      core.MEM_OUT16(addr, core.registers.map[inst.reg2].value & 0xFFFF)

    else:
      core.MEM_OUT8(addr, core.registers.map[inst.reg2].value & 0xFF)

  @staticmethod
  def jit(core, inst):
    reg1 = core.registers.map[inst.reg1]
    reg2 = core.registers.map[inst.reg2]

    if inst.opcode == DuckyOpcodes.STW:
      writer = core.MEM_OUT32

      if inst.immediate_flag == 1:
        offset = inst.sign_extend_immediate(core.LOGGER, inst)

        if offset == 0:
          def __jit_stw():
            writer(reg1.value, reg2.value)

          return __jit_stw

        else:
          def __jit_stw():
            writer((reg1.value + offset) % 4294967296, reg2.value)

          return __jit_stw

      else:
        def __jit_stw():
          writer(reg1.value, reg2.value)

        return __jit_stw

    elif inst.opcode == DuckyOpcodes.STS:
      writer = core.MEM_OUT16

      if inst.immediate_flag == 1:
        offset = inst.sign_extend_immediate(core.LOGGER, inst)

        if offset == 0:
          def __jit_sts():
            writer(reg1.value, reg2.value)

          return __jit_sts

        else:
          def __jit_sts():
            writer((reg1.value + offset) % 4294967296, reg2.value)

          return __jit_sts

      else:
        def __jit_sts():
          writer(reg1.value, reg2.value)

        return __jit_sts

    elif inst.opcode == DuckyOpcodes.STB:
      writer = core.MEM_OUT8

      if inst.immediate_flag == 1:
        offset = inst.sign_extend_immediate(core.LOGGER, inst)

        if offset == 0:
          def __jit_stb():
            writer(reg1.value, reg2.value & 0xFF)

          return __jit_stb

        else:
          def __jit_stb():
            writer((reg1.value + offset) % 4294967296, reg2.value & 0xFF)

          return __jit_stb

      else:
        def __jit_stb():
          writer(reg1.value, reg2.value & 0xFF)

        return __jit_stb

    return None

class _LOAD_IMM(Descriptor_R_I):
  @classmethod
  def load(cls, core, inst):
    raise NotImplementedError('%s does not implement "load immediate" method' % cls.__name__)

  @classmethod
  def execute(cls, core, inst):
    cls.load(core, inst)
    update_arith_flags(core, core.registers.map[inst.reg])

class LW(_LOAD):
  mnemonic = 'lw'
  opcode = DuckyOpcodes.LW

class LS(_LOAD):
  mnemonic = 'ls'
  opcode = DuckyOpcodes.LS

class LB(_LOAD):
  mnemonic = 'lb'
  opcode = DuckyOpcodes.LB

class LI(_LOAD_IMM):
  mnemonic = 'li'
  opcode   = DuckyOpcodes.LI

  @classmethod
  def load(cls, core, inst):
    core.registers.map[inst.reg].value = inst.sign_extend_immediate(core.LOGGER, inst)

  @staticmethod
  def jit(core, inst):
    reg = core.registers.map[inst.reg]
    i = inst.sign_extend_immediate(core.LOGGER, inst)

    if i == 0:
      def __jit_li():
        reg.value = 0
        core.arith_zero = True
        core.arith_overflow = False
        core.arith_sign = False

      return __jit_li

    else:
      sign = (i & 0x80000000) != 0

      def __jit_li():
        reg.value = i
        core.arith_zero = False
        core.arith_overflow = False
        core.arith_sign = sign

      return __jit_li

class LIU(_LOAD_IMM):
  mnemonic = 'liu'
  opcode   = DuckyOpcodes.LIU

  @classmethod
  def load(cls, core, inst):
    r = core.registers.map[inst.reg]

    r.value = (r.value & 0xFFFF) | ((inst.sign_extend_immediate(core.LOGGER, inst) & 0xFFFF) << 16)

class LA(_LOAD_IMM):
  mnemonic = 'la'
  opcode   = DuckyOpcodes.LA
  relative_address = True

  @classmethod
  def load(cls, core, inst):
    core.registers.map[inst.reg].value = core.registers.ip.value + inst.sign_extend_immediate(core.LOGGER, inst)

  @staticmethod
  def jit(core, inst):
    reg = core.registers.map[inst.reg]
    ip = core.registers.ip
    offset = inst.sign_extend_immediate(core.LOGGER, inst)

    if offset == 0:
      def __jit_la():
        reg.value = v = ip.value
        core.arith_zero = v == 0
        core.arith_overflow = False
        core.arith_sign = (v & 0x80000000) != 0

      return __jit_la

    else:
      def __jit_la():
        reg.value = v = ip.value + offset
        core.arith_zero = v == 0
        core.arith_overflow = False
        core.arith_sign = (v & 0x80000000) != 0

      return __jit_la

class STW(_STORE):
  mnemonic = 'stw'
  opcode   = DuckyOpcodes.STW

class STS(_STORE):
  mnemonic = 'sts'
  opcode   = DuckyOpcodes.STS

class STB(_STORE):
  mnemonic = 'stb'
  opcode   = DuckyOpcodes.STB

class MOV(Descriptor_R_R):
  mnemonic = 'mov'
  opcode = DuckyOpcodes.MOV
  encoding = EncodingR

  @staticmethod
  def execute(core, inst):
    core.registers.map[inst.reg1].value = core.registers.map[inst.reg2].value

  @staticmethod
  def jit(core, inst):
    reg1 = core.registers.map[inst.reg1]
    reg2 = core.registers.map[inst.reg2]

    def __jit_mov():
      reg1.value = reg2.value

    return __jit_mov

class SWP(Descriptor_R_R):
  mnemonic = 'swp'
  opcode = DuckyOpcodes.SWP
  encoding = EncodingR

  @staticmethod
  def execute(core, inst):
    r1 = core.registers.map[inst.reg1]
    r2 = core.registers.map[inst.reg2]

    r1.value, r2.value = r2.value, r1.value

  @staticmethod
  def jit(core, inst):
    reg1 = core.registers.map[inst.reg1]
    reg2 = core.registers.map[inst.reg2]

    def __jit_swp():
      reg1.value, reg2.value = reg2.value, reg1.value

    return __jit_swp

#
# Control instructions
#
class CTR(Descriptor_R_R):
  mnemonic = 'ctr'
  opcode = DuckyOpcodes.CTR
  encoding = EncodingR

  @staticmethod
  def execute(core, inst):
    core.registers.map[inst.reg1].value = core.control_coprocessor.read(inst.reg2)
    update_arith_flags(core, core.registers.map[inst.reg1])

class CTW(Descriptor_R_R):
  mnemonic = 'ctw'
  opcode = DuckyOpcodes.CTW
  encoding = EncodingR

  @staticmethod
  def execute(core, inst):
    core.control_coprocessor.write(inst.reg1, core.registers.map[inst.reg2].value)

class FPTC(Descriptor):
  mnemonic = 'fptc'
  opcode = DuckyOpcodes.FPTC
  encoding = EncodingI

  @staticmethod
  def execute(core, inst):
    core.mmu.release_ptes()


class DuckyInstructionSet(InstructionSet):
  instruction_set_id = 0
  opcodes = DuckyOpcodes

NOP(DuckyInstructionSet)
INT(DuckyInstructionSet)
IPI(DuckyInstructionSet)
RETINT(DuckyInstructionSet)
CALL(DuckyInstructionSet)
RET(DuckyInstructionSet)
CLI(DuckyInstructionSet)
STI(DuckyInstructionSet)
HLT(DuckyInstructionSet)
RST(DuckyInstructionSet)
IDLE(DuckyInstructionSet)
PUSH(DuckyInstructionSet)
POP(DuckyInstructionSet)
INC(DuckyInstructionSet)
DEC(DuckyInstructionSet)
ADD(DuckyInstructionSet)
SUB(DuckyInstructionSet)
CMP(DuckyInstructionSet)
J(DuckyInstructionSet)
INW(DuckyInstructionSet)
INB(DuckyInstructionSet)
INS(DuckyInstructionSet)
OUTS(DuckyInstructionSet)
OUTW(DuckyInstructionSet)
OUTB(DuckyInstructionSet)
AND(DuckyInstructionSet)
OR(DuckyInstructionSet)
XOR(DuckyInstructionSet)
NOT(DuckyInstructionSet)
SHIFTL(DuckyInstructionSet)
SHIFTR(DuckyInstructionSet)

# Memory access
LW(DuckyInstructionSet)
LS(DuckyInstructionSet)
LB(DuckyInstructionSet)
LI(DuckyInstructionSet)
LIU(DuckyInstructionSet)
LA(DuckyInstructionSet)
STW(DuckyInstructionSet)
STS(DuckyInstructionSet)
STB(DuckyInstructionSet)

MOV(DuckyInstructionSet)
SWP(DuckyInstructionSet)
MUL(DuckyInstructionSet)
DIV(DuckyInstructionSet)
UDIV(DuckyInstructionSet)
MOD(DuckyInstructionSet)
CMPU(DuckyInstructionSet)
CAS(DuckyInstructionSet)
SIS(DuckyInstructionSet)

# Branching instructions
BE(DuckyInstructionSet)
BNE(DuckyInstructionSet)
BZ(DuckyInstructionSet)
BNZ(DuckyInstructionSet)
BO(DuckyInstructionSet)
BNO(DuckyInstructionSet)
BS(DuckyInstructionSet)
BNS(DuckyInstructionSet)
BG(DuckyInstructionSet)
BGE(DuckyInstructionSet)
BL(DuckyInstructionSet)
BLE(DuckyInstructionSet)

# SET* instructions
SETE(DuckyInstructionSet)
SETNE(DuckyInstructionSet)
SETZ(DuckyInstructionSet)
SETNZ(DuckyInstructionSet)
SETO(DuckyInstructionSet)
SETNO(DuckyInstructionSet)
SETS(DuckyInstructionSet)
SETNS(DuckyInstructionSet)
SETG(DuckyInstructionSet)
SETGE(DuckyInstructionSet)
SETL(DuckyInstructionSet)
SETLE(DuckyInstructionSet)

LPM(DuckyInstructionSet)

# Control instructions
CTR(DuckyInstructionSet)
CTW(DuckyInstructionSet)
FPTC(DuckyInstructionSet)

DuckyInstructionSet.init()

INSTRUCTION_SETS = {
  DuckyInstructionSet.instruction_set_id: DuckyInstructionSet
}

def get_instruction_set(i, exc = None):
  from ..cpu import InvalidInstructionSetError
  exc = exc or InvalidInstructionSetError

  if i not in INSTRUCTION_SETS:
    raise exc(i)

  return INSTRUCTION_SETS[i]



import sys
import ctypes
import optparse
import re
import tabulate

from six import viewkeys

from . import add_common_options, parse_options
from ..cpu.instructions import DuckyInstructionSet, get_instruction_set
from ..mm import u16_t, u32_t, UINT16_FMT, SIZE_FMT, UINT32_FMT, UINT8_FMT
from ..mm.binary import File, SectionTypes, SECTION_TYPES, SYMBOL_DATA_TYPES, SymbolDataTypes, RelocFlags, SymbolFlags, SectionFlags

from ..cpu.coprocessor.math_copro import MathCoprocessorInstructionSet  # noqa

def show_file_header(logger, f):
  f_header = f.get_header()

  logger.info('=== File header ===')
  logger.info('  Magic:    0x%X', f_header.magic)
  logger.info('  Version:  %i', f_header.version)
  logger.info('  Sections: %i', f_header.sections)
  logger.info('  Flags:    %s (0x%02X)', ''.join(['M' if f_header.flags.mmapable == 1 else '-']), ctypes.cast(ctypes.byref(f_header.flags), ctypes.POINTER(ctypes.c_ubyte)).contents.value)
  logger.info('')

def show_sections(logger, f):
  logger.info('=== Sections ===')
  logger.info('')

  f_header = f.get_header()

  table = [
    ['Index', 'Name', 'Type', 'Flags', 'Base', 'Items', 'Data size', 'File size', 'Offset']
  ]

  for i in range(0, f_header.sections):
    header, content = f.get_section(i)

    header_flags = SectionFlags.from_encoding(header.flags)

    table.append([
      header.index,
      f.string_table.get_string(header.name),
      SECTION_TYPES[header.type],
      '%s (%s)' % (header_flags.to_string(), UINT16_FMT(header_flags.to_int())),
      UINT32_FMT(header.base),
      header.items,
      SIZE_FMT(header.data_size),
      SIZE_FMT(header.file_size),
      SIZE_FMT(header.offset)
    ])

  logger.table(table)

  logger.info('')

def show_disassemble(logger, f):
  logger.info('=== Disassemble ==')
  logger.info('')

  instruction_set = DuckyInstructionSet
  f_header = f.get_header()

  for i in range(0, f_header.sections):
    header, content = f.get_section(i)

    if header.type != SectionTypes.TEXT:
      continue

    logger.info('  Section %s', f.string_table.get_string(header.name))

    csp = header.base
    for raw_inst in content:
      csp_str = UINT32_FMT(csp)
      csp += 4

      inst, desc, opcode = instruction_set.decode_instruction(logger, raw_inst.value)
      logger.info('  %s (%s) %s', csp_str, UINT32_FMT(raw_inst), instruction_set.disassemble_instruction(logger, raw_inst.value))

      if opcode == DuckyInstructionSet.opcodes.SIS:
        instruction_set = get_instruction_set(inst.immediate)

  logger.info('')

def show_reloc(logger, f):
  logger.info('=== Reloc entries ===')
  logger.info('')

  f_header = f.get_header()

  table = [
    ['Name', 'Flags', 'Patch section', 'Patch address', 'Patch offset', 'Patch size']
  ]

  for i in range(0, f_header.sections):
    header, content = f.get_section(i)

    if header.type != SectionTypes.RELOC:
      continue

    for index, entry in enumerate(content):
      _header, _content = f.get_section(entry.patch_section)

      table.append([
        f.string_table.get_string(entry.name),
        RelocFlags.from_encoding(entry.flags).to_string(),
        f.string_table.get_string(_header.name),
        UINT32_FMT(entry.patch_address),
        entry.patch_offset,
        entry.patch_size
      ])

  logger.table(table)
  logger.info('')

def show_symbols(logger, f):
  ascii_replacements = {
    '\n':   '\\n',
    '\r':   '\\r',
    '\x00': '\\0'
  }

  def ascii_replacer(m):
    return ascii_replacements[m.group(0)]

  ascii_replace = re.compile(r'|'.join(viewkeys(ascii_replacements)))

  logger.info('=== Symbols ===')
  logger.info('')

  f_header = f.get_header()

  table = [
    ['Name', 'Section', 'Flags', 'Address', 'Type', 'Size', 'File', 'Line', 'Content']
  ]

  for i in range(0, f_header.sections):
    header, content = f.get_section(i)

    if header.type != SectionTypes.SYMBOLS:
      continue

    for index, entry in enumerate(content):
      _header, _content = f.get_section(entry.section)

      table_row = [
        f.string_table.get_string(entry.name),
        f.string_table.get_string(_header.name),
        SymbolFlags.from_encoding(entry.flags).to_string(),
        UINT32_FMT(entry.address),
        '%s (%i)' % (SYMBOL_DATA_TYPES[entry.type], entry.type),
        SIZE_FMT(entry.size),
        f.string_table.get_string(entry.filename),
        entry.lineno
      ]

      symbol_content = ''

      if _header.flags.bss == 1:
        pass

      else:
        if entry.type == SymbolDataTypes.INT:
          def __get(i):
            return _content[entry.address - _header.base + i].value << (8 * i)

          symbol_content = u32_t(__get(0) | __get(1) | __get(2) | __get(3))
          symbol_content = UINT32_FMT(symbol_content.value)

        elif entry.type == SymbolDataTypes.SHORT:
          def __get(i):
            return _content[entry.address - _header.base + i].value << (8 * i)

          symbol_content = u16_t(__get(0) | __get(1))
          symbol_content = UINT16_FMT(symbol_content.value)

        elif entry.type in (SymbolDataTypes.BYTE, SymbolDataTypes.CHAR):
          symbol_content = UINT8_FMT(_content[entry.address - _header.base].value)

        elif entry.type == SymbolDataTypes.ASCII:
          symbol_content = ''.join(['%s' % chr(c.value) for c in _content[entry.address - _header.base:entry.address - _header.base + entry.size]])

        elif entry.type == SymbolDataTypes.STRING:
          symbol_content = ''.join(['%s' % chr(c.value) for c in _content[entry.address - _header.base:entry.address - _header.base + entry.size]])

        if entry.type == SymbolDataTypes.ASCII or entry.type == SymbolDataTypes.STRING:
          if len(symbol_content) > 32:
            symbol_content = symbol_content[0:29] + '...'

          symbol_content = '"' + ascii_replace.sub(ascii_replacer, symbol_content) + '"'

      table_row.append(symbol_content)
      table.append(table_row)

  for line in tabulate.tabulate(table, headers = 'firstrow', tablefmt = 'simple', numalign = 'right').split('\n'):
    logger.info(line)

  logger.info('')

def main():
  parser = optparse.OptionParser()
  add_common_options(parser)

  parser.add_option('-i', dest = 'file_in', action = 'append', default = [], help = 'File to inspect')

  parser.add_option('-H', dest = 'header',      default = False, action = 'store_true', help = 'Show file header')
  parser.add_option('-D', dest = 'disassemble', default = False, action = 'store_true', help = 'Disassemble TEXT sections')
  parser.add_option('-s', dest = 'symbols',     default = False, action = 'store_true', help = 'List symbols')
  parser.add_option('-r', dest = 'reloc',       default = False, action  ='store_true', help = 'List reloc entries')
  parser.add_option('-S', dest = 'sections',    default = False, action = 'store_true', help = 'List sections')
  parser.add_option('-a', dest = 'all',         default = False, action = 'store_true', help = 'All of above')

  options, logger = parse_options(parser)

  if not options.file_in:
    parser.print_help()
    sys.exit(1)

  if options.all:
    options.header = options.disassemble = options.symbols = options.sections = options.reloc = True

  for file_in in options.file_in:
    logger.info('Input file: %s', file_in)

    with File.open(logger, file_in, 'r') as f_in:
      f_in.load()

      logger.info('')

      if options.header:
        show_file_header(logger, f_in)

      if options.sections:
        show_sections(logger, f_in)

      if options.symbols:
        show_symbols(logger, f_in)

      if options.reloc:
        show_reloc(logger, f_in)

      if options.disassemble:
        show_disassemble(logger, f_in)



import os
import sys
import traceback

from pycparser import c_ast

from ..import dump_node, show_node

PASSES = sys.modules[__name__]
PASSES_DIR = os.path.dirname(PASSES.__file__)


AST_PASSES = {}
BT_PASSES  = {}


class ASTVisitor(c_ast.NodeVisitor):
  priority = 99

  def __init__(self, logger, *args, **kwargs):
    super(ASTVisitor, self).__init__(*args, **kwargs)

    self.logger = logger
    self.DEBUG = logger.debug
    self.INFO = logger.info
    self.WARN = logger.warning
    self.ERROR = logger.error

    self._log_prefix_stack = []
    self.log_prefix = ''

    self.node_parent = None

    self.tree_modified = False

  def UP(self):
    self.log_prefix = (len(self.log_prefix) - 8) * ' '

    if traceback.extract_stack()[-3][2] != self._log_prefix_stack[-1]:
      self.ERROR('whops! tos=%s, current=%s', self._log_prefix_stack[-1], traceback.extract_stack()[-3][2])

    del self._log_prefix_stack[-1]

  def DOWN(self):
    self._log_prefix_stack.append(traceback.extract_stack()[-3][2])
    self.log_prefix += 8 * ' '

  def generic_visit(self, node):
    old_parent = self.node_parent
    self.node_parent = node

    l = [self.visit(c) for c_name, c in node.children()]

    self.node_parent = old_parent

    return l

  def visit(self, node, **kwargs):
    method = 'visit_' + node.__class__.__name__
    visitor = getattr(self, method, self.generic_visit)

    return visitor(node, **kwargs)

class BlockVisitor(object):
  priority = 99

  def __init__(self, logger, *args, **kwargs):
    self.logger = logger
    self.DEBUG = logger.debug
    self.INFO = logger.info
    self.WARN = logger.warning
    self.ERROR = logger.error

    self.log_prefix = ''

    self.tree_modified = False

  def UP(self):
    self.log_prefix = (len(self.log_prefix) - 8) * ' '

  def DOWN(self):
    self.log_prefix += 8 * ' '

  def do_visit_block(self, block):
    pass

  def visit_block(self, block):
    self.DEBUG(self.log_prefix + 'Examine block %s', block.id)
    self.DOWN()

    self.do_visit_block(block)

    self.UP()

  def do_visit_fn(self, fn):
    for block in fn.blocks:
      self.visit_block(block)

  def visit_fn(self, fn):
    self.DEBUG(self.log_prefix + 'Examine fn %s', fn.name)
    self.DOWN()

    self.do_visit_fn(fn)

    self.UP()

  def do_visit(self, cv):
    for fn in cv.functions:
      self.visit_fn(fn)

  def visit(self, cv):
    self.DEBUG(self.log_prefix + '%s pass', self.__class__.__name__)
    self.DOWN()

    self.do_visit(cv)

    self.UP()

class ASTOptVisitor(ASTVisitor):
  def replace_child(self, current_node, new_node):
    self.DEBUG(self.log_prefix + 'PRE:')
    for line in show_node(self.node_parent).split('\n'):
      self.DEBUG(self.log_prefix + line)

    for name, child in self.node_parent.children():
      if child != current_node:
        continue

      setattr(self.node_parent, name, new_node)
      self.tree_modified = True
      self.DEBUG(self.log_prefix + 'Replaced by %s', dump_node(new_node))
      break

    self.DEBUG(self.log_prefix + 'POST:')
    for line in show_node(self.node_parent).split('\n'):
      self.DEBUG(self.log_prefix + line)

def load(logger):
  passes = {}

  for filename in os.listdir(PASSES_DIR):
    if not filename.endswith(".py") or filename.startswith("_"):
      continue

    if not os.path.isfile(os.path.join(PASSES_DIR, filename)):
      continue

    pass_module = filename[:-3]

    try:
      __import__(PASSES.__name__, {}, {}, [pass_module])
      passes[pass_module] = sys.modules[PASSES.__name__ + '.' + pass_module]
      logger.debug('Pass %s loaded', pass_module)

    except (ImportError, SyntaxError):
      logger.exception('Failed to load pass %s', pass_module)

  return passes



import colorama
import logging
import tabulate

from six import print_

LEVELS = {
  logging.DEBUG:    'DEBG',
  logging.INFO:     'INFO',
  logging.WARNING:  'WARN',
  logging.ERROR:    'ERRR',
  logging.CRITICAL: 'CRIT'
}

COLORS = {
  logging.DEBUG:    colorama.Fore.WHITE,
  logging.INFO:     colorama.Fore.GREEN,
  logging.WARNING:  colorama.Fore.YELLOW,
  logging.ERROR:    colorama.Fore.RED,
  logging.CRITICAL: colorama.Fore.RED
}

COLOR_RESET = colorama.Fore.RESET + colorama.Back.RESET + colorama.Style.RESET_ALL

def RED(s):
  return colorama.Fore.RED + s + COLOR_RESET

def GREEN(s):
  return colorama.Fore.GREEN + s + COLOR_RESET

def BLUE(s):
  return colorama.Fore.BLUE + s + COLOR_RESET

def WHITE(s):
  return colorama.Fore.WHITE + s + COLOR_RESET

class LogFormatter(logging.Formatter):
  def format(self, record):
    try:
      prefix = '{color_start}{stamp:.02f} [{level}] '.format(color_start = COLORS[record.levelno], level = LEVELS[record.levelno], stamp = record.created)
      postfix = COLOR_RESET

      msg = [prefix + record.getMessage() + postfix]

      if record.exc_info is not None:
        msg += [prefix + l + postfix for l in self.formatException(record.exc_info).split('\n')[-15:]]

      return '\n'.join(msg)

    except Exception:
      import sys
      import traceback
      print_('Failure in formatter:', file = sys.stderr)
      print_('record: ' + str(record), file = sys.stderr)
      print_('message: ' + str(record.msg), file = sys.stderr)
      print_('args: ' + str(record.args), file = sys.stderr)
      traceback.print_exc(file = sys.stderr)
      sys.exit(1)

class StreamHandler(logging.StreamHandler):
  def __init__(self, *args, **kwargs):
    super(StreamHandler, self).__init__(*args, **kwargs)

    self.setFormatter(LogFormatter())

def create_logger(name = None, handler = None):
  name = name or 'ducky'

  logger = logging.getLogger(name)
  if handler:
    logger.addHandler(handler)

  def __table(table, fn = None, **kwargs):
    fn = fn or logger.info

    for line in tabulate.tabulate(table, headers = 'firstrow', tablefmt = 'simple', numalign = 'right', **kwargs).split('\n'):
      fn(line)

  logger.table = __table

  return logger



"""
Hardware Description Table structures.
"""

from ctypes import LittleEndianStructure, sizeof
from enum import IntEnum

from .mm import u8_t, u16_t, u32_t
from .util import str2int

#: Magic number present in HDT header
HDT_MAGIC = 0x4D5E6F70

class HDTEntryTypes(IntEnum):
  """
  Types of different HDT entries.
  """

  UNDEFINED = 0
  CPU       = 1
  MEMORY    = 2
  ARGUMENT  = 3

class HDTStructure(LittleEndianStructure):
  """
  Base class of all HDT structures.
  """

  _pack_ = 0

class HDTHeader(HDTStructure):
  """
  HDT header. Contains magic number, number of HDT entries that immediately
  follow header.
  """

  _fields_ = [
    ('magic',   u32_t),
    ('entries', u32_t),
    ('length',  u32_t)
  ]

  def __init__(self):
    HDTStructure.__init__(self)

    self.magic = HDT_MAGIC
    self.entries = 0

class HDTEntry(HDTStructure):
  """
  Base class of all HDT entries.

  Each entry has at least two fields, `type` and `length`. Other fields
  depend on type of entry, and follow immediately after `length` field.

  :param u16_t type: type of entry. See :py:class:`ducky.hdt.HDTEntryTypes`.
  :param u16_t length: length of entry, in bytes.
  """

  def __init__(self, entry_type, length):
    HDTStructure.__init__(self)

    self.type = entry_type
    self.length = length

  @classmethod
  def create(cls, *args, **kwargs):
    return [cls(*args, **kwargs)]

class HDTEntry_CPU(HDTEntry):
  """
  HDT entry describing CPU configuration.

  :param u16_t nr_cpus: number of CPUs.
  :param u16_t nr_cores: number of cores per CPU.
  """

  _fields_ = [
    ('type',     u16_t),
    ('length',   u16_t),
    ('nr_cpus',  u16_t),
    ('nr_cores', u16_t)
  ]

  def __init__(self, logger, config):
    HDTEntry.__init__(self, HDTEntryTypes.CPU, sizeof(HDTEntry_CPU))

    self.nr_cpus = config.getint('machine', 'cpus')
    self.nr_cores = config.getint('machine', 'cores')

    logger.debug('HDTEntry_CPU: nr_cpus=%s, nr_cores=%s', self.nr_cpus, self.nr_cores)

class HDTEntry_Memory(HDTEntry):
  """
  HDT entry describing memory configuration.

  :param u32_t size: size of memory, in bytes.
  """

  _fields_ = [
    ('type',   u16_t),
    ('length', u16_t),
    ('size',   u32_t)
  ]

  def __init__(self, logger, config):
    HDTEntry.__init__(self, HDTEntryTypes.MEMORY, sizeof(HDTEntry_Memory))

    self.size = config.getint('memory', 'size', 0x1000000)

    logger.debug('HDTEntry_Memory: size=%s', self.size)

class HDTEntry_Argument(HDTEntry):
  """
  """

  _fields_ = [
    ('type',         u16_t),
    ('length',       u16_t),
    ('name_length',  u8_t),
    ('value_length', u8_t),
    ('name',         u8_t * 13),
    ('value',        u8_t * 13)
  ]

  MAX_NAME_LENGTH = 13

  def __init__(self, arg_name, arg_type, arg_value):
    HDTEntry.__init__(self, HDTEntryTypes.ARGUMENT, sizeof(HDTEntry_Argument))

    def __encode_string(field, s):
      for i, c in enumerate(s):
        getattr(self, field)[i] = ord(c)

        if i == HDTEntry_Argument.MAX_NAME_LENGTH - 1:
          break

      setattr(self, field + '_length', i + 1)

    __encode_string('name', arg_name)

    if arg_type == 'int':
      arg_value = str2int(arg_value)

      self.value[0] = arg_value & 0xFF
      self.value[1] = (arg_value >> 8) & 0xFF
      self.value[2] = (arg_value >> 16) & 0xFF
      self.value[3] = (arg_value >> 24) & 0xFF

    else:
      __encode_string('value', arg_value)

  @classmethod
  def create(cls, logger, config):
    if not config.has_section('arguments'):
      return []

    arguments = []

    for arg_name in config.options('arguments'):
      arg_type, arg_value = config.get('arguments', arg_name).split(',')
      arg_type = arg_type.strip()
      arg_value = arg_value.strip()

      arguments.append(HDTEntry_Argument(arg_name, arg_type, arg_value))

    return arguments

class HDT(object):
  """
  Root of HDT. Provides methods for creating HDT for a given machine configuration.

  :param logging.Logger logger: logger instance used for logging.
  :param ducky.config.MachineConfig config: configuration file HDT should reflect.
  """

  klasses = [
    HDTEntry_Memory,
    HDTEntry_CPU,
    HDTEntry_Argument
  ]

  def __init__(self, logger, config = None):
    self.logger = logger
    self.config = config

    self.header = None
    self.entries = []

  def __len__(self):
    """
    Get size of HDT - sum of entries' lengths and length of a header.

    :rtype: int
    :returns: size of HDT, in bytes.
    """

    return sizeof(HDTHeader) + sum([sizeof(entry) for entry in self.entries])

  def create(self):
    """
    Create HDT header and entries from config file.
    """

    self.header = HDTHeader()

    for klass in HDT.klasses:
      self.entries += klass.create(self.logger, self.config)

    self.header.entries = len(self.entries)
    self.header.length = len(self)



import curses
import logging
import sys
import time

from curtsies import fsarray, fmtstr, FullscreenWindow
from curtsies.window import FIRST_COLUMN
from functools import partial
from six import itervalues

RED = partial(fmtstr, fg = 'red')
GREEN = partial(fmtstr, fg = 'green')
BLUE = partial(fmtstr, fg = 'blue')
GRAY = partial(fmtstr, fg = 'gray')


COLOR_DISABLED = GRAY
COLOR_ENABLED  = GREEN
COLOR_FOCUS    = RED

class Widget(object):
  def __init__(self, width = None, height = None, left = None, top = None, enabled = False, visible = True, logger = None):
    super(Widget, self).__init__()

    self.logger = logger or logging.getLogger()

    self.logger.debug('%s.__init__: width=%d, height=%d, left=%d, top=%d, enabled=%s, visible=%s', self.__class__.__name__, width, height, left, top, enabled, visible)

    assert width is not None
    assert height is not None
    assert left is not None
    assert top is not None

    self.width  = width
    self.height = height
    self.left   = left
    self.top    = top

    self.handle = None

    self._enabled      = enabled
    self._visible      = visible
    self._needs_redraw = False
    self._focus        = False

  def on_needs_redraw(self, value):
    self._needs_redraw = value
    return True

  def on_focus(self, value):
    self._focus = value
    self.needs_redraw = True
    return True

  def on_enabled(self, value):
    self._enabled = value
    self.needs_redraw = True
    return True

  def on_visible(self, value):
    self._visible = value
    if value is True:
      self.needs_redraw = True
    return True

  needs_redraw = property(lambda self: self._needs_redraw, lambda self, value: self.on_event(('needs_redraw', value)))
  focus = property(lambda self: self._focus, lambda self, value: self.on_event(('focus', value)))
  enabled = property(lambda self: self._enabled, lambda self, value: self.on_event(('enabled', value)))
  visible = property(lambda self: self._visible, lambda self, value: self.on_event(('visible', value)))

  def attach_handle(self, handle):
    self.handle = handle

  def _deliver_local_event(self, e):
    if isinstance(e, tuple):
      name, value = e
      return getattr(self, 'on_%s' % name)(value)

    return None

  def on_event(self, e):
    if self._deliver_local_event(e) is True:
      return True

    handler = 'on_%s' % e
    if not hasattr(self, handler):
      return False

    return getattr(self, handler)(e)

  def clear(self):
    self.logger.debug('%s.clear', self.__class__.__name__)

    self.handle.clear()

  def erase(self):
    self.logger.debug('%s.erase', self.__class__.__name__)

    self.handle.erase()

  def border_color(self):
    self.logger.debug('%s.border_color: focus=%s, enabled=%s', self.__class__.__name__, self.focus, self.enabled)

    return curses.color_pair(2 if self.focus is True else (1 if self.enabled is True else 0))

  def apply_border(self):
    self.logger.debug('%s.apply_border: %s %s', self.__class__.__name__, self.handle, self.border_color())

    self.handle.attron(self.border_color())
    self.handle.border('|', '|', '-', '-', '+', '+', '+', '+')
    self.handle.attroff(self.border_color())

  def do_draw(self):
    raise NotImplementedError()

  def draw(self):
    self.logger.debug('%s.draw', self.__class__.__name__)

    self.do_draw()
    self.apply_border()
    self.handle.noutrefresh()
    self.needs_redraw = False

  def do_update(self, model):
    raise NotImplementedError()

  def update(self, model):
    self.logger.debug('%s.draw', self.__class__.__name__)

    if self.visible is not True:
      return

    self.do_update(model)

    if self.needs_redraw is not True:
      return

    self.draw()

class Container(Widget):
  def __init__(self, *args, **kwargs):
    super(Container, self).__init__(*args, **kwargs)

    self._children = []

  def __getitem__(self, key):
    return self._children[key]

  def __setitem__(self, key, item):
    self._children[key] = item
    self.dirty = True

  def __delitem__(self, key):
    del self._children[key]
    self.dirty = True

  def __iter__(self):
    i = 0
    try:
      while True:
        v = self[i]
        yield v
        i += 1
    except IndexError:
      return

  def append(self, item):
    self._children.append(item)

  def remove(self, item):
    self._children.remove(item) 

  def on_event(self, e):
    if self._deliver_local_event(e) is True:
      return True

    for widget in self:
      if widget.focus is not True:
        continue

      return widget.on_event(e)

    return None

class Window(Container):
  def __init__(self, title = None, **kwargs):
    super(Window, self).__init__(**kwargs)

    assert title is not None

    self.title = title

  def __repr__(self):
    return '<%s: title="%s">' % (self.__class__.__name__, self.title)

  def apply_border(self):
    super(Window, self).apply_border()

    self.handle.addstr(0, 2, ' %s ' % self.title)

class Button(Widget):
  def __init__(self, width, label, on_click = None, **kwargs):
    super(Button, self).__init__(width, 3, **kwargs)

    self.label = label
    self.on_click = on_click

  def __repr__(self):
    return '<%s: %s>' % (self.__class__.__name__, self.label)

  def on_Enter(self, e):
    if self.on_click is not None:
      self.on_click(e)

    return True

  def do_update(self, model):
    return True

  def do_draw(self):
    self.handle.addstr(self.top + 2, self.left + 2, self.label)

class Page(Container):
  def __init__(self, name = None, *args, **kwargs):
    kwargs['width'] = 0
    kwargs['height'] = 0
    kwargs['left'] = 0
    kwargs['top'] = 0

    super(Page, self).__init__(*args, **kwargs)

    assert name is not None

    self.name = name

  def __repr__(self):
    return '<%s: name=%s, widgets=[%s]>' % (self.__class__.__name__, self.name, ', '.join([repr(widget) for widget in self]))

class Screen(Container):
  def __init__(self, width = None, height = None, logger = None):
    super(Screen, self).__init__(width = width, height = height, left = 0, top = 0, logger = logger)

    self.pages = {}
    self.current_page = None

  def __repr__(self):
    return '<%s>' % self.__class__.__name__

  def add_page(self, page):
    self.logger.debug('%s.add_page: page=%r', self.__class__.__name__, page)

    assert page.name not in self.pages

    self.pages[page.name] = page

  def set_page(self, name):
    self.logger.debug('%s.set_page: name=%s', self.__class__.__name__, name)

    assert name in self.pages

    self.current_page = self.pages[name]

  def commit(self):
    self.logger.debug('%s.commit', self.__class__.__name__)

    curses.doupdate()

  def on_event(self, e):
    self.logger.debug('%s.on_event: e=%r', self.__class__.__name__, e)

    if e == '<Ctrl-j>':
      e = 'Enter'

    if self._deliver_local_event(e) is True:
      return True

    v = False

    if self.current_page is not None:
      v = self.current_page.on_event(e)

    if v is not True:
      print >> sys.stderr, 'Unhandled event: e=%s' % e

  def create_attrs(self, color = 0, blink = False, bold = False):
    attrs = curses.color_pair(color)

    if blink is True:
      attrs |= curses.A_BLINK

    if bold is True:
      attrs |= curses.A_BOLD

    return attrs

def render_loop(logger, screen, event_quit, delay = 0.1):
  import fcntl
  import os
  import sys

  #from ..streams import fd_blocking
  #fd_blocking(sys.stdout.fileno(), block = True)

  def loop(stdscr):
    # Setup colors
    curses.use_default_colors()

    for i in range(0, curses.COLORS):
      curses.init_pair(i, i, -1);

    screen.attach_handle(stdscr)

    for page in itervalues(screen.pages):
      for widget in page:
        widget.attach_handle(curses.newwin(widget.height, widget.width, widget.top, widget.left))

    while not event_quit.is_set():
      screen.update()
      screen.commit()
      #time.sleep(delay)

    screen.erase()
    screen.commit()

  curses.wrapper(loop)

def event_loop(logger, screen, event_quit):
  import sys
  from curtsies import Input

  with Input(keynames='curtsies') as input_generator:
    for e in Input():
      logger.debug('event_loop: event=%s', repr(e))

      screen.on_event(e)

      if event_quit.is_set():
        logger.debug('event_loop: quit')
        break
